/*
 *
 *   (C) Copyright IBM Corp. 2001, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: object.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "fullengine.h"
#include "object.h"
#include "engine.h"
#include "handlemgr.h"
#include "discover.h"
#include "common.h"
#include "volume.h"
#include "internalAPI.h"
#include "message.h"
#include "memman.h"
#include "common.h"


int isa_valid_input_object(ADDRESS object,
			   TAG     object_tag,
			   ADDRESS object_handle,
			   ADDRESS parameters) {
	int rc = 0;

	storage_object_t * obj = (storage_object_t *) object;
	storage_container_t * disk_group = (storage_container_t *) parameters;

	LOG_PROC_ENTRY();

	if ((object_tag == EVMS_OBJECT_TAG) ||
	    (object_tag == REGION_TAG) ||
	    (object_tag == SEGMENT_TAG) ||
	    (object_tag == DISK_TAG)) {

		/* The object must not be corrupt. */
		if (!(obj->flags & SOFLAG_CORRUPT)) {

			/* The object must be a top level object. */
			if (is_top_object(obj)) {

				/* The object must not insist on being the top object. */
				if (!(obj->flags & SOFLAG_MUST_BE_TOP)) {

					/* The object must be in the specified disk group. */
					if (obj->disk_group == disk_group) {
						LOG_DEBUG("Object %s is a valid input object.\n", obj->name);

					} else {
						LOG_ERROR("Object %s in disk group %s is not in disk group %s.\n", obj->name, (obj->disk_group != NULL) ? obj->disk_group->name : "(local)", (disk_group != NULL) ? disk_group->name : "(local)");
						rc = EINVAL;
					}

				} else {
					LOG_ERROR("Object %s is not a valid input object.  It insists it must be a top level object.\n", obj->name);
					rc = EINVAL;
				}

			} else {
				LOG_ERROR("Object %s is not a valid input object.  It is not a top level object.\n", obj->name);
				rc = EINVAL;
			}

		} else {
			LOG_ERROR("Object %s is not a valid input object.  It is corrupt.\n", obj->name);
			rc = EINVAL;
		}

	} else {
		switch (object_tag) {
		case PLUGIN_TAG:
			LOG_ERROR("Object is a plug-in.\n");
			break;
		case CONTAINER_TAG:
			LOG_ERROR("Object is a container.\n");
			break;
		case VOLUME_TAG:
			LOG_ERROR("Object is a volume.\n");
			break;
		default:
			LOG_ERROR("Object is of unknown type %ld.\n", object_tag);
			break;
		}
		rc = EINVAL;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int validate_create_parameters(plugin_handle_t     plugin_handle,
				      handle_array_t   *  objects,
				      plugin_record_t * * ppPlugRec,
				      dlist_t *           pObjectList) {
	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	*ppPlugRec = NULL;
	*pObjectList = NULL;

	rc = translate_handle(plugin_handle,
			      &object,
			      &type);

	if (rc == HANDLE_MANAGER_NO_ERROR) {
		if (type == PLUGIN) {
			plugin_record_t * pPlugRec = (plugin_record_t *) object;
			plugin_type_t plugin_type = GetPluginType(pPlugRec->id);

			*ppPlugRec = pPlugRec;

			if ((plugin_type == EVMS_DEVICE_MANAGER) ||
			    (plugin_type == EVMS_SEGMENT_MANAGER) ||
			    (plugin_type == EVMS_REGION_MANAGER) ||
			    (plugin_type == EVMS_FEATURE) ||
			    (plugin_type == EVMS_ASSOCIATIVE_FEATURE)) {
				dlist_t object_list = CreateList();

				if (object_list != NULL) {

					rc = make_dlist(objects, object_list);

					if (rc == 0) {
						TAG tag;
						storage_object_t * obj;
						storage_container_t * disk_group = NULL;

						if (BlindGetObject(object_list, &tag,NULL,FALSE,(ADDRESS *) &obj) == DLIST_SUCCESS) {
							disk_group = obj->disk_group;
						}

						/*
						 * Make sure that the input object list contains
						 * top level objects and none of them insist on
						 * being a top object.
						 */
						rc = ForEachItem(object_list,
								 isa_valid_input_object,
								 disk_group,
								 TRUE);
					}

					if (rc != DLIST_SUCCESS) {
						DestroyList(&object_list, FALSE);
					}

					*pObjectList = object_list;

				} else {
					LOG_CRITICAL("Error allocating memory for an output object list.\n");
					rc = ENOMEM;
				}

			} else {
				LOG_ERROR("The plug-in %s is not a type that manages storage objects.\n", pPlugRec->short_name);
				rc = EINVAL;
			}

		} else {
			LOG_ERROR("The plugin_handle is not for a plug-in.\n");
			rc = EINVAL;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int evms_create(plugin_handle_t    plugin_handle,
		handle_array_t   * objects,
		option_array_t   * options,
		handle_array_t * * output_objects) {

	int rc = 0;
	plugin_record_t * pPlugRec = NULL;
	dlist_t object_list = NULL;
	dlist_t input_objects;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc != 0) {
		LOG_PROC_EXIT_INT(rc);
		return rc;
	}

	rc = validate_create_parameters(plugin_handle, objects,
					&pPlugRec,     &object_list);
	if (rc == 0) {
		/*
		 * Make a copy of the input objects list in case we have to
		 * clean up after an error.
		 */
		input_objects = CreateList();

		if (input_objects != NULL) {
			rc = CopyList(input_objects, object_list, AppendToList);

			if (rc == DLIST_SUCCESS) {
				/*
				 * Make a list to receive the new objects from the
				 * create.
				 */
				dlist_t new_objects = CreateList();

				if (new_objects != NULL) {

					/*
					 * We are finally setup to do the create.
					 */
					rc = pPlugRec->functions.plugin->create(object_list, options, new_objects);

					if (rc == 0) {

						int tmp_rc;
						storage_object_t * new_obj;
						storage_object_t * child_obj;

						/* Remove any corrupt objects from the list. */
						PruneList(new_objects,
							  remove_corrupt_object,
							  NULL);

						DLIST_FOR_EACH(new_obj, new_objects, tmp_rc) {
							int tmp_rc2;

							/*
							 * Now that the input objects are
							 * consumed, the parent can write
							 * on them anywhere, including over
							 * any stop data.  Forget that the
							 * child objects have stop data.
							 */
							DLIST_FOR_EACH(child_obj, new_obj->child_objects, tmp_rc2) {
								child_obj->flags &= ~SOFLAG_HAS_STOP_DATA;
							}

							/* Mark the new objects' feature
							 * headers dirty.
							 */
							set_feature_header_dirty(new_obj,
										 get_tag_for_object(new_obj),
										 NULL,
										 NULL);
						}

						propigate_cluster_info(new_objects);

						/*
						 * Make sure the compatibility volume names
						 * are in sync.
						 */
						sync_volumes();

						/* Sort the list that was affected. */
						switch (GetPluginType(pPlugRec->id)) {
							case EVMS_DEVICE_MANAGER:
								SortList(DiskList, compare_objects);
								break;
							case EVMS_SEGMENT_MANAGER:
								SortList(SegmentList, compare_objects);
								break;
							case EVMS_REGION_MANAGER:
								SortList(RegionList, compare_objects);
								break;
							case EVMS_FEATURE:
							case EVMS_ASSOCIATIVE_FEATURE:
								SortList(EVMSObjectList, compare_objects);
								break;
							default:
								break;
						}

						/*
						 * If the user wants handles for the
						 * new objects, make the handles.
						 */
						if (output_objects != NULL) {
							rc = make_user_handle_array(new_objects, output_objects);
						}
					}

					DestroyList(&new_objects, FALSE);

				} else {
					LOG_CRITICAL("Error getting memory for resulting object list.\n");
					rc = ENOMEM;
				}

			} else {
				LOG_CRITICAL("Error code %d from CopyList when making a copy of the input object list.\n", rc);
			}

			DestroyList(&input_objects, FALSE);

		} else {
			LOG_CRITICAL("Error getting memory for copying the input object list.\n");
			rc = ENOMEM;
		}

		DestroyList(&object_list, FALSE);
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * validate_transfer_parameters is a support function for evms_transfer()
 * below.  It does the parameter validation so that evms_transfer() can
 * do the real work of transferring the object.
 */
static int validate_transfer_parameters(object_handle_t         object_handle,
					plugin_handle_t         plugin_handle,
					object_handle_t         container_handle,
					storage_object_t    * * p_object,
					plugin_record_t     * * p_plugin,
					storage_container_t * * p_container) {
	int rc = 0;
	object_type_t         type;
	storage_object_t    * object = NULL;
	plugin_record_t     * plugin = NULL;
	storage_container_t * container = NULL;

	LOG_PROC_ENTRY();

	rc = translate_handle(object_handle, (void *) &object, &type);

	if (rc == HANDLE_MANAGER_NO_ERROR) {

		/* Containers can only contain storage objects. */
		if ((type == DISK) ||
		    (type == SEGMENT) ||
		    (type == REGION) ||
		    (type == EVMS_OBJECT)) {

			/* The object must not be part of a volume. */
			if (object->volume != NULL) {
				LOG_ERROR("The object cannot be transferred since it is part of a volume.\n");
				rc = EINVAL;
			}

		} else {
			LOG_ERROR("The object handle is not for a storage object.\n");
			rc = EINVAL;
		}

	} else {
		LOG_ERROR("The object handle parameter is not valid.  translate_handle() returned error %d.\n", rc);
	}

	if (rc == 0) {
		/* If a plugin_handle was given, get the plug-in record. */
		if (plugin_handle != 0) {
			rc = translate_handle(plugin_handle, (void *) &plugin, &type);

			if (rc == 0) {
				if (type == PLUGIN) {
					/*
					 * Check to make sure the plug-in supports containers, i.e.,
					 * it has a function table for container functions.
					 */
					if (plugin->container_functions == NULL) {
						LOG_ERROR("The target plug-in does not support container functions.\n");
						rc = EINVAL;
					}

				} else {
					LOG_ERROR("The plug-in handle parameter is not a handle for a plug-in.\n");
					rc = EINVAL;
				}

			} else {
				LOG_ERROR("The plug-in handle parameter is not valid.  translate_handle() returned error %d.\n", rc);
			}
		}
	}

	if (rc == 0) {
		/* If a container_handle was given, get the container. */
		if (container_handle != 0) {
			rc = translate_handle(container_handle, (void *) &container, &type);

			if (rc == 0) {
				if (type != CONTAINER) {
					LOG_ERROR("The container handle parameter is not a handle for a container.\n");
					rc = EINVAL;
				}

			} else {
				LOG_ERROR("The container handle parameter is not valid.  translate_handle() returned error %d.\n", rc);
			}
		}
	}

	if (rc == 0) {
		/*
		 * If both a plug-in and a container were given, make sure the plug-in
		 * is the one that manages the container.
		 */
		if ((plugin != NULL) && (container != NULL)) {
		    if (container->plugin != plugin) {
			LOG_ERROR("Parameter conflict.  The specified container is not managed by the specified plug-in.\n");
			rc = EINVAL;
		    }
		}
	}

	if (rc == 0) {
		/*
		 * If a container was given, make sure the object and the
		 * container are in the same disk group.
		 */
		if (container != NULL) {
			if (object->disk_group != container->disk_group) {
				LOG_ERROR("Object %s and container %s are not in the same disk group.\n", object->name, container->name);
				rc = EINVAL;
			}
		}
	}

	/* If all went well, return the pointers. */
	if (rc == 0) {
		*p_object    = object;
		*p_plugin    = plugin;
		*p_container = container;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * evms_transfer will transfer an object to a particular plug-in and/or
 * container.  If container is 0, the plug-in will assign the object to its
 * default container.
 */
int evms_transfer(object_handle_t  object_handle,
		  plugin_handle_t  plugin_handle,
		  object_handle_t  container_handle,
		  option_array_t * options) {

	int rc = 0;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		storage_object_t    * object = NULL;
		plugin_record_t     * plugin = NULL;
		storage_container_t * container = NULL;

		rc = validate_transfer_parameters(object_handle, plugin_handle, container_handle,
						  &object, &plugin, &container);
		if (rc == 0) {
			/*
			 * If a target plug-in was not specified, then the target plug-in is
			 * the one that manages the target container, if a container was
			 * specified.
			 */
			if (plugin == NULL) {
				if (container != NULL) {
					plugin = container->plugin;
				}
			}

			/* Check if the object is currently in a container. */
			if (object->consuming_container != NULL) {

				if (object->consuming_container->plugin == plugin) {

					/* This is an intra-plugin transfer. */
					rc = plugin->container_functions->transfer_object(object, container, options);

					LOG_DEBUG("transfer_object() returned error code %d.\n", rc);

				} else {
					/*
					 * This is an inter-plugin transfer.  Tell the plug-in that
					 * currently owns the container to which the object belongs
					 * to remove it from its current container.
					 */
					rc = object->consuming_container->plugin->container_functions->remove_object(object);

					LOG_DEBUG("remove_object() returned error code %d.\n", rc);

					if (rc == 0) {
						/*
						 * If a target plug-in/container was specified, add the
						 * object to the target plug-in/container.
						 */
						if (plugin != NULL) {
							rc = plugin->container_functions->add_object(object, container, options);

							LOG_DEBUG("add_object() returned error code %d.\n", rc);
						}

					} else {
						LOG_WARNING("The %s plug-in failed to remove the object with handle %d from its current container.  The error code was %d.\n", object->consuming_container->plugin->short_name,object_handle, rc);
					}
				}

			} else {
				if (plugin != NULL) {
					/*
					 * The object does not belong to a container.
					 * Just add it to the new container.
					 */
					rc = plugin->container_functions->add_object(object, container, options);
					LOG_DEBUG("add_object() returned error code %d.\n", rc);

					if (rc == 0) {
						/*
						 * Now that the object is consumed,
						 * any stop data on it can be
						 * overwritten.  Forget that it is
						 * there.
						 */
						object->flags &= ~SOFLAG_HAS_STOP_DATA;
					}
				} else {
					LOG_WARNING("The object does not belong to a container and no target container or plug-in was specified.  There is nothing for evms_transfer() to do.\n");
					rc = EINVAL;
				}
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * evms_get_object_list returns a pointer to a handle_array_t with handles for
 * storage objects, optionally filtering on the object type, the data type of
 * the object, the plug-in that manages the object, the cluster node on which
 * the object resides, and some flags.  If any of the filtering parameters is 0,
 * that filter is not used.
 */
int evms_get_object_list(object_type_t         object_type,
			 data_type_t           data_type,
			 plugin_handle_t       plugin_handle,
			 object_handle_t       disk_group_handle,
			 object_search_flags_t flags,
			 handle_array_t    * * object_handle_list) {
	int rc = 0;
	ADDRESS object = NULL;
	object_type_t type;
	plugin_record_t * plugin = NULL;
	storage_container_t * disk_group = NULL;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();

	if (rc == 0) {
		if (plugin_handle != 0) {
			/* Translate the handle for the feature to make sure it is valid */
			/* and to get the plugin_record_t for the feature. */
			rc = translate_handle(plugin_handle,
					      &object,
					      &type);

			if (rc == HANDLE_MANAGER_NO_ERROR) {
				if (type == PLUGIN) {
					plugin = (plugin_record_t *) object;
				} else {
					rc = EINVAL;
				}
			}
		}

		if (rc == 0) {
			if (disk_group_handle != 0) {
				/*
				 * Translate the handle for the disk group
				 * to make sure it is valid and to get the
				 * disk group storage_container_t.
				 */
				rc = translate_handle(disk_group_handle,
						      &object,
						      &type);

				if (rc == HANDLE_MANAGER_NO_ERROR) {
					if (type == CONTAINER) {
						disk_group = (storage_container_t *) object;
					} else {
						rc = EINVAL;
					}
				}
			}
		}

		if (rc == 0) {
			dlist_t object_list;

			/*
			 * Call the internal version of engine_get_object_list.  "plugin"
			 * will be NULL if the caller did not specify a plug-in, else it
			 * will be a pointer to the plug-in's plugin_record_t.
			 */
			rc = engine_get_object_list(object_type, data_type, plugin, disk_group, flags, &object_list);

			if (rc == 0) {
				rc = make_user_handle_array(object_list, object_handle_list);

				/* We are finished with the list that was returned by */
				/* engine_get_object_list. */
				DestroyList(&object_list, FALSE);
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * evms_assign
 */
int evms_assign(object_handle_t  object_handle,
		plugin_handle_t  plugin_handle,
		option_array_t * options) {

	int rc = 0;
	ADDRESS object = NULL;
	object_type_t type;
	storage_object_t * obj = NULL;
	plugin_record_t * plugin = NULL;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {

		if (object_handle != 0) {
			/*
			 * Translate the handle for the object to make sure it is valid
			 * and to get the storage_object_t for the object.
			 */
			rc = translate_handle(object_handle,
					      &object,
					      &type);

			if (rc == HANDLE_MANAGER_NO_ERROR) {
				if ((type == DISK) ||
				    (type == SEGMENT) ||
				    (type == REGION) ||
				    (type == EVMS_OBJECT)) {
					storage_object_t * parent;
					TAG tag;

					obj = (storage_object_t *) object;
					rc = BlindGetObject(obj->parent_objects,
							    &tag,
							    NULL,
							    FALSE,
							    (ADDRESS *) &parent);

					if (rc == DLIST_SUCCESS) {
						LOG_ERROR("Object %s already has parent objects produced by plug-in %s.\n", obj->name, parent->plugin->short_name);
					}

				} else {
					LOG_ERROR("Handle %d is not a handle for a storage object.\n", object_handle);
					rc = EINVAL;
				}
			}

		} else {
			LOG_ERROR("An object handle must be given.\n");
			rc = EINVAL;
		}

		if (plugin_handle != 0) {
			/*
			 * Translate the handle for the plug-in to make sure it is valid
			 * and to get the plugin_record_t for the plug-in.
			 */
			rc = translate_handle(plugin_handle,
					      &object,
					      &type);

			if (rc == HANDLE_MANAGER_NO_ERROR) {
				if (type == PLUGIN) {
					plugin = (plugin_record_t *) object;
				} else {
					LOG_ERROR("Handle %d is not a handle for a plug-in.\n", plugin_handle);
					rc = EINVAL;
				}
			}

		} else {
			LOG_ERROR("An plug-in handle must be given.\n");
			rc = EINVAL;
		}

		if (rc == 0) {
			rc = plugin->functions.plugin->assign(obj, options);

			if (rc == 0) {
				/*
				 * Now that the object is consumed,
				 * any stop data on it can be
				 * overwritten.  Forget that it is
				 * there.
				 */
				obj->flags &= ~SOFLAG_HAS_STOP_DATA;

				propigate_cluster_info(obj->parent_objects);
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return(rc);
}


static int get_plugin(ADDRESS object,
		      TAG     object_tag,
		      ADDRESS object_handle,
		      ADDRESS parameters) {

	int rc = DLIST_SUCCESS;
	storage_object_t * obj = (storage_object_t *) object;
	plugin_record_t  * * pp_plugin_record = (plugin_record_t * *) parameters;

	LOG_PROC_ENTRY();

	if (*pp_plugin_record == NULL) {
		*pp_plugin_record = obj->plugin;

	} else {
		if (*pp_plugin_record != obj->plugin) {
			LOG_ERROR("Object %s does not have parent objects that are all managed by the same plug-in.\n", obj->name);
			rc = EINVAL;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int get_parent_plugin(storage_object_t * obj, plugin_record_t * * p_parent_plugin) {

	int rc = 0;

	LOG_PROC_ENTRY();

	if (!ListEmpty(obj->parent_objects)) {
		rc = ForEachItem(obj->parent_objects,
				 get_plugin,
				 p_parent_plugin,
				 TRUE);

	} else {
		LOG_ERROR("Object %s does not have any parent objects.\n", obj->name);
		rc = EINVAL;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int is_in_use(ADDRESS object,
		     TAG     object_tag,
		     ADDRESS object_handle,
		     ADDRESS parameters) {

	int rc = DLIST_SUCCESS;
	storage_object_t * obj = (storage_object_t *) object;

	LOG_PROC_ENTRY();

	if (!ListEmpty(obj->parent_objects)) {
		LOG_ERROR("Object %s has parent objects.\n", obj->name);
		rc = EINVAL;
	}

	if (obj->consuming_container != NULL) {
		LOG_ERROR("Object %s is part of a container.\n", obj->name);
		rc = EINVAL;
	}

	if (obj->volume != NULL) {
		LOG_ERROR("Object %s is used as volume %s.\n", obj->name, obj->volume->name);
		rc = EINVAL;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * evms_can_unassign determines if a plug-in can be unassigned from consuming
 * an object.  The parent objects must themselves not have parent objects,
 * must not be consumed by a container, and if a volume is created from them,
 * the volume must not be mounted.  If that is true, then ask the plug-in
 * if it can unassign itself from the object.
 */
int evms_can_unassign(object_handle_t object_handle) {

	int rc = 0;
	ADDRESS object = NULL;
	object_type_t type;
	storage_object_t * obj = NULL;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {

		if (object_handle != 0) {
			/*
			 * Translate the handle for the object to make sure it is valid
			 * and to get the storage_object_t for the object.
			 */
			rc = translate_handle(object_handle,
					      &object,
					      &type);

			if (rc == HANDLE_MANAGER_NO_ERROR) {
				if ((type == DISK) ||
				    (type == SEGMENT) ||
				    (type == REGION) ||
				    (type == EVMS_OBJECT)) {
					obj = (storage_object_t *) object;
				} else {
					LOG_DETAILS("Handle %d is not a handle for a storage object.\n", object_handle);
					rc = EINVAL;
				}
			}

		} else {
			LOG_ERROR("An object handle must be given.\n");
			rc = EINVAL;
		}

		if (rc == 0) {
			/*
			 * Check to make sure the parent objects don't have parents
			 * themselves, that they are not part of a container, and
			 * that they are not part of a volume.
			 */
			rc = ForEachItem(obj->parent_objects,
					 is_in_use,
					 NULL,
					 TRUE);

			if (rc == DLIST_SUCCESS) {
				plugin_record_t * parent_plugin = NULL;

				/*
				 * Get the plug-in that is managing the parents of this
				 * object.
				 */
				rc = get_parent_plugin(obj, &parent_plugin);

				if (rc == 0) {
					rc = parent_plugin->functions.plugin->can_unassign(obj);
				}
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return(rc);
}


static int is_data_object(ADDRESS object,
			  TAG     object_tag,
			  ADDRESS object_handle,
			  ADDRESS parameters) {

	int rc = DLIST_SUCCESS;
	storage_object_t * obj = (storage_object_t *) object;

	LOG_PROC_ENTRY();

	if (obj->data_type == DATA_TYPE) {
		rc = EEXIST;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * evms_unassign unassigns a plug-in from consuming an object.  This is most
 * useful for segment managers that claim a disk or segment.
 */
int evms_unassign(object_handle_t object_handle) {

	int rc = 0;
	ADDRESS object = NULL;
	object_type_t type;
	storage_object_t * obj = NULL;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {

		if (object_handle != 0) {
			/*
			 * Translate the handle for the object to make sure it is valid
			 * and to get the storage_object_t for the object.
			 */
			rc = translate_handle(object_handle,
					      &object,
					      &type);

			if (rc == HANDLE_MANAGER_NO_ERROR) {
				if ((type == DISK) ||
				    (type == SEGMENT) ||
				    (type == REGION) ||
				    (type == EVMS_OBJECT)) {
					obj = (storage_object_t *) object;
				} else {
					LOG_ERROR("Handle %d is not a handle for a storage object.\n", object_handle);
					rc = EINVAL;
				}
			}

		} else {
			LOG_ERROR("An object handle must be given.\n");
			rc = EINVAL;
		}

		if (rc == 0) {
			/*
			 * Check to make sure the parent objects don't have parents
			 * themselves, that they are not part of a container, and
			 * that they are not part of a volume.
			 */
			rc = ForEachItem(obj->parent_objects,
					 is_in_use,
					 NULL,
					 TRUE);

			if (rc == DLIST_SUCCESS) {
				plugin_record_t * parent_plugin = NULL;

				/*
				 * Get the plug-in that is managing the parents of this
				 * object.
				 */
				rc = get_parent_plugin(obj, &parent_plugin);

				if (rc == 0) {
					/*
					 * If this object has parent objects that are
					 * data objects, warn the user that the data
					 * objects will be destroyed.
					 */
					rc = ForEachItem(obj->parent_objects,
							 is_data_object,
							 NULL,
							 TRUE);

					if (rc != 0) {
						char * choices[] = {"Continue", "Cancel", NULL};
						int answer = 0;	    /* Default is "Continue" */

						rc = engine_user_message(&answer, choices,
									 "WARNING:  Plug-in %s is producing data objects from object %s.  "
									 "Unassigning plug-in %s from object %s will destroy the data objects.\n",
									 parent_plugin->short_name, obj->name,
									 parent_plugin->short_name, obj->name);

						if (answer == 1) {
							/* User wants to cancel. */
							rc = E_CANCELED;
						}
					}

					if (rc == 0) {
						rc = parent_plugin->functions.plugin->unassign(obj);

						if (rc == 0) {
							/*
							 * The previous owner may
							 * have trashed stop data while the
							 * object was consumed.
							 */
							obj->flags &= ~SOFLAG_HAS_STOP_DATA;
						}
					}
				}
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return(rc);
}


static void free_function_info_array_contents(void * object) {

	int i;
	function_info_array_t * function_array = (function_info_array_t *) object;

	for (i = 0; i < function_array->count; i++) {
		if (function_array->info[i].title != NULL) {
			engine_free(function_array->info[i].title);
		}
		if (function_array->info[i].verb != NULL) {
			engine_free(function_array->info[i].verb);
		}
		if (function_array->info[i].help != NULL) {
			engine_free(function_array->info[i].help);
		}
	}
}

/*
 * Get an array of private actions that are supported by the plug-in that is
 * managing "thing".
 */
int evms_get_plugin_functions(engine_handle_t           thing_handle,
			      function_info_array_t * * actions) {

	int rc = 0;
	ADDRESS object = NULL;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();

	if (rc == 0) {

		if (thing_handle != 0) {
			rc = translate_handle(thing_handle,
					      &object,
					      &type);

			if (rc == HANDLE_MANAGER_NO_ERROR) {
				function_info_array_t * function_array = NULL;
				switch (type) {
				case PLUGIN:
					{
						plugin_record_t * plugin = (plugin_record_t *) object;

						switch (GetPluginType(plugin->id)) {
						case EVMS_DEVICE_MANAGER:
						case EVMS_SEGMENT_MANAGER:
						case EVMS_REGION_MANAGER:
						case EVMS_FEATURE:
						case EVMS_ASSOCIATIVE_FEATURE:
							rc = plugin->functions.plugin->get_plugin_functions(NULL, &function_array);
							break;

						case EVMS_FILESYSTEM_INTERFACE_MODULE:
							rc = plugin->functions.fsim->get_plugin_functions(NULL, &function_array);
							break;

						case EVMS_CLUSTER_MANAGER_INTERFACE_MODULE:
							rc = plugin->functions.cluster->get_plugin_functions(NULL, &function_array);
							break;

						default:
							LOG_ERROR("Plug-in %s has an unknown type of %#x.  Can't get plug-in functions for that type of plug-in.\n", plugin->short_name, GetPluginType(plugin->id));
							rc = ENOSYS;
							break;
						}
					}
					break;

				case EVMS_OBJECT:
				case REGION:
				case SEGMENT:
				case DISK:
					{
						storage_object_t * obj = (storage_object_t *) object;

						rc = obj->plugin->functions.plugin->get_plugin_functions(obj, &function_array);
					}
					break;

				case CONTAINER:
					{
						storage_container_t * con = (storage_container_t *) object;

						rc = con->plugin->container_functions->get_plugin_functions(con, &function_array);
					}
					break;

				case VOLUME:
					{
						logical_volume_t * vol = (logical_volume_t *) object;

						if (vol->file_system_manager != NULL) {
							rc = vol->file_system_manager->functions.fsim->get_plugin_functions(vol, &function_array);

						} else {
							rc = ENOSYS;
						}
					}
					break;

				default:
					LOG_ERROR("Handle %d is of type %d which is not valid for this function.\n", thing_handle, type);
					rc = EINVAL;
				}

				if (rc == 0) {
					/*
					 * Allocate a user memory block and copy the action info
					 * to the user memory.
					 */
					uint size = sizeof(function_info_array_t);
					if (function_array->count > 1) {
						size += (function_array->count - 1) * sizeof(function_info_t);
					}

					*actions = alloc_app_struct(size, free_function_info_array_contents);

					if (*actions != NULL) {
						/* We do not copy the strings, just the pointers to the strings. */
						memcpy(*actions, function_array, size);
					} else {
						rc = ENOMEM;
					}

					/*
					 * Free the action array returned by the plug-in but don't
					 * free the strings.  The user copy now has the pointers to the strings.
					 * the strings will be freed when the user frees the user copy of
					 * the function_info_array_t.
					 */
					engine_free(function_array);
				}
			}

		} else {
			LOG_ERROR("A handle must be given.\n");
			rc = EINVAL;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Execute a plug-in private action on a "thing".
 */
int evms_do_plugin_function(engine_handle_t  thing_handle,
			    task_action_t    action,
			    handle_array_t * objects,
			    option_array_t * options) {

	int rc = 0;
	ADDRESS object = NULL;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {

		if (thing_handle != 0) {
			rc = translate_handle(thing_handle,
					      &object,
					      &type);

			if (rc == HANDLE_MANAGER_NO_ERROR) {
				dlist_t object_list = CreateList();

				if (object_list != NULL) {

					rc = make_dlist(objects, object_list);

					if (rc == 0) {

						switch (type) {
						case PLUGIN:
							{
								plugin_record_t * plugin = (plugin_record_t *) object;

								if (GetPluginType(plugin->id) == EVMS_FILESYSTEM_INTERFACE_MODULE) {
									rc = plugin->functions.fsim->plugin_function(NULL, action, object_list, options);
								} else {
									rc = plugin->functions.plugin->plugin_function(NULL, action, object_list, options);
								}
							}
							break;

						case EVMS_OBJECT:
						case REGION:
						case SEGMENT:
						case DISK:
							{
								storage_object_t * obj = (storage_object_t *) object;

								rc = obj->plugin->functions.plugin->plugin_function(obj, action, object_list, options);
							}
							break;

						case CONTAINER:
							{
								storage_container_t * con = (storage_container_t *) object;

								rc = con->plugin->container_functions->plugin_function(con, action, object_list, options);
							}
							break;

						case VOLUME:
							{
								logical_volume_t * vol = (logical_volume_t *) object;

								if (vol->file_system_manager != NULL) {
									rc = vol->file_system_manager->functions.fsim->plugin_function(vol, action, object_list, options);

								} else {
									rc = ENOSYS;
								}
							}
							break;

						default:
							LOG_ERROR("Handle %d is of type %d which is not valid for this function.\n", thing_handle, type);
							rc = EINVAL;
						}
					}

					DestroyList(&object_list, FALSE);

				} else {
					LOG_CRITICAL("Error allocating memory for an output object list.\n");
					rc = ENOMEM;
				}
			}

		} else {
			LOG_ERROR("A handle must be given.\n");
			rc = EINVAL;
		}
	}

	if (rc == 0) {
		/*
		 * A plug-in function can add or remove child objects to/from
		 * the given object.  Since we don't know what went on inside
		 * the set_info() and how it affected other objects, update the
		 * stop data state on all the objects that can have stop data.
		 */
		update_all_stop_data_states();
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

