/*
 *   (C) Copyright IBM Corp. 2001, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: AIX Region Manager
 * File: evms2/engine/plugins/aix/aix_options.c
 *
 * Description: This file contains functions related to parsing and verifying
 *		option arrays and setting up and manipulating option
 *		descriptors for engine tasks.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "aixregmgr.h"


/********** Container Creation Options **********/


/* aix_create_container_parse_options
 *
 *	Examine the option array for container creation options. Current valid
 *	options are:
 *	PP_SIZE
 *
 *	The "pp_size" argument to this function is in vsectors, NOT in
 *	exponent form.
 */
int aix_create_container_parse_options( option_array_t  * options,
					unsigned long   * pp_size )
{
	unsigned int    i;
	int             rc;

	LOG_ENTRY();

	// Default PP size is 4MB.
	*pp_size = AIX_DEFAULT_PP_SIZE;

	for ( i = 0; i < options->count; i++ ) {
		// If only the name-based index is specified, get the number.
		if ( ! options->option[i].is_number_based ) {
			if ( ! strcmp(options->option[i].name, AIX_OPTION_CONTAINER_PP_SIZE_STR) ) {
				options->option[i].number = AIX_OPTION_CONTAINER_PP_SIZE_INDEX;
			}
			else {
				continue;
			}
		}

		LOG_EXTRA("Parsing option %d\n", options->option[i].number);

		// Use the number-based index to record the option.
		switch ( options->option[i].number ) {
		case AIX_OPTION_CONTAINER_PP_SIZE_INDEX:
			*pp_size = options->option[i].value.ui32;
			LOG_DEBUG("Setting PP size: %ld sectors\n", *pp_size);
			break;
		default:
			break;
		}
	}

	rc = aix_create_container_verify_options(pp_size);

	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_create_container_verify_options
 *
 *	Check each of the options to make sure they are valid for creating
 *	a new container.
 *
 *	The "pp_size" argument to this function is in vsectors, NOT in
 *	exponent form.
 */
int aix_create_container_verify_options( unsigned long * pp_size )
{
	int rc;

	LOG_ENTRY();

	rc = aix_check_pp_size(pp_size);

	LOG_EXIT_INT(rc);
	return rc;
}


/* Function: aix_create_container_allocate_options
 *
 *	This is a helper function for aix_init_options. It sets up the initial
 *	information in the option descriptor for a Create_Container task.
 */
int aix_create_container_allocate_option_descriptor( option_desc_array_t * od )
{
	LOG_ENTRY();

	od->count = AIX_OPTION_CONTAINER_PP_SIZE_INDEX + 1;

	// Option 0 is PE size
	SET_STRING(od->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].name, AIX_OPTION_CONTAINER_PP_SIZE_STR);
	SET_STRING(od->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].title, "PP size for new container");
	SET_STRING(od->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].tip, "Acceptable range: 1MB to 1GB. Must be a power of 2");
	od->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].type     = EVMS_Type_Unsigned_Int32;
	od->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].unit     = EVMS_Unit_Sectors;
	od->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].min_len  = 0;
	od->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].max_len  = 0;
	od->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].flags    = EVMS_OPTION_FLAGS_NOT_REQUIRED|EVMS_OPTION_FLAGS_AUTOMATIC;
	od->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].constraint_type = EVMS_Collection_List;
	SET_POWER2_LIST(od->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].constraint.list, AIX_MIN_PP_SIZE, AIX_MAX_PP_SIZE);
	od->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].value.ui32 = AIX_DEFAULT_PP_SIZE;

	LOG_EXIT_INT(0);
	return 0;
}


/* Function: aix_create_container_get_acceptable
 *
 *	This is a helper function for aix_init_task. It determines which
 *	objects in EVMS are appropriate for a Create_Container task.
 */
int aix_create_container_get_acceptable( dlist_t acceptable_objects )
{
	storage_object_t        * object;
	dlist_t                 object_list;
	sector_count_t          num_sectors;
	int                     rc;

	LOG_ENTRY();

	// Find all top-level disks, segments, and regions in the system. This
	// will also retrieve AIX regions, so let the user beware. :)
	rc = EngFncs->get_object_list(DISK|SEGMENT|REGION, DATA_TYPE, 0, NULL, TOPMOST, &object_list);
	if (rc) {
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Don't add objects that are too small or too big to use with AIX.
	DLIST_FOR_EACH(object, object_list, rc) {
		num_sectors = AVAILABLE_PP_SECTORS(object);
		LOG_DEBUG("Checking object->name:%s num_sectors:%"PRIu64" \n", object->name, num_sectors);
		if ( num_sectors > AIX_MIN_PP_SIZE &&
		     num_sectors / AIX_MAX_PP_SIZE <= AIX_MAX_PPS_PER_PV ) {
			aix_add_object_to_list(object, acceptable_objects, AppendToList);
		}
		else{
			LOG_DEBUG("throwing out object:%s \n",object->name);
		}
	}

	DestroyList(&object_list, FALSE);

	LOG_EXIT_INT(0);
	return 0;
}



/* Function: aix_create_container_set_objects
 *
 *	Examine the selected_objects list in the task, and verify that each
 *	object is valid for inclusion in a new container. If any are not valid,
 *	remove them from the selected_objects and add them to the
 *	declined_objects. In create_container, no options will change based
 *	on which objects are selected.
 */
int aix_create_container_set_objects(   task_context_t  * context,
					dlist_t         declined_objects,
					task_effect_t   * effect )
{
	storage_object_t        * object;
	unsigned long           pp_size = context->option_descriptors->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].value.ui32;
	int                     rc;

	LOG_ENTRY();

	// Check that all of the objects are valid for use in a new container.
	DLIST_FOR_EACH(object, context->selected_objects, rc) {

		rc = aix_check_object_for_container_inclusion(object);
		if (rc) {
// Add this object to declined objects.
			LOG_ERROR("One or more objects are invalid for container creation.\n");
			LOG_EXIT_INT(EINVAL);
			return EINVAL;
		}

		// If any of the segments are too small for the default PE
		// size, try to reset the PE size.
		rc = aix_check_object_for_pp_size(object, &pp_size);
		if (rc == EINVAL) {
// Add this object to declined objects.
			LOG_ERROR("One or more objects are invalid for container creation.\n");
			LOG_EXIT_INT(EINVAL);
			return EINVAL;
		}
		else if (rc) {
			LOG_DEBUG("%s is too small for default PP size.\n", object->name);
			LOG_DEBUG("Resetting initial PP size value to %ld sectors.\n", pp_size);
			context->option_descriptors->option[AIX_OPTION_CONTAINER_PP_SIZE_INDEX].value.ui32 = pp_size;
			*effect |= EVMS_Effect_Reload_Options;
		}
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* Function: aix_create_container_set_option
 *
 *	This is a helper function for aix_set_option. It sets one particular
 *	option in the option descriptor for a Create_Container task, and
 *	updates other option descriptor information accordingly.
 */
int aix_create_container_set_option(    task_context_t  * context,
					u_int32_t       index,
					value_t         * value,
					task_effect_t   * effect )
{
	option_desc_array_t     * od = context->option_descriptors;
	storage_object_t        * object;
	unsigned long           pp_size;
	int                     rc = 0;

	LOG_ENTRY();
	LOG_EXTRA("Setting option %d\n", index);

	switch (index) {
	
	case AIX_OPTION_CONTAINER_PP_SIZE_INDEX:
		pp_size = value->ui32;

		rc = aix_check_pp_size(&pp_size);
		if (rc) {
			*effect |= EVMS_Effect_Inexact;
		}

		// Make sure every object in the "selected" list is large
		// enough for the specified pp_size.
		DLIST_FOR_EACH(object, context->selected_objects, rc) {
			rc = aix_check_object_for_pp_size(object, &pp_size);
			if (rc) {
				LOG_ERROR("One or more objects too small for PE size %d.\n", value->ui32);
				rc = ENOSPC;
				break;
			}
		}

		if ( rc != ENOSPC ) {
			od->option[index].value.ui32 = pp_size;
			od->option[index].flags &= ~EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
			rc = 0;
		}
		break;

	default:
		rc = EINVAL;
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}



/********** Region Creation Options **********/



/* aix_create_region_parse_options
 *
 *	Examine the option array for region creation options. Current valid
 *	options are:
 *	NAME
 *	NUM_LPS
 */
int aix_create_region_parse_options(    option_array_t          * options,
					aix_region_options_t    * region_opts )
{
	unsigned int    i;
	int             rc;

	LOG_ENTRY();

	//memset(region_opts, 0, sizeof(aix_region_options_t));

	for ( i = 0; i < options->count; i++ ) {
		// If only the name-based index is specified, get the number.
		if ( ! options->option[i].is_number_based ) {
			if ( ! strcmp(options->option[i].name, AIX_OPTION_REGION_NAME_STR) ) {
				options->option[i].number = AIX_OPTION_REGION_NAME_INDEX;
			}
			else if ( ! strcmp(options->option[i].name, AIX_OPTION_REGION_NUM_LPS_STR) ) {
				options->option[i].number = AIX_OPTION_REGION_NUM_LPS_INDEX;
			}
			else {
				continue;
			}
		}

		LOG_EXTRA("Parsing option %d\n", options->option[i].number);

		// Use the number-based index to record the option.
		switch ( options->option[i].number ) {
		case AIX_OPTION_REGION_NAME_INDEX:
			strncpy(region_opts->name, options->option[i].value.s, AIX_NAMESIZ-1);
			LOG_DEBUG("Setting name: %s\n", region_opts->name);
			break;
		case AIX_OPTION_REGION_NUM_LPS_INDEX:
			region_opts->num_lps = options->option[i].value.ui32;
			LOG_DEBUG("Setting number of LPs: %ld\n", region_opts->num_lps);
			break;
		default:
			break;
		}
	}

	rc = aix_create_region_verify_options(region_opts);

	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_create_region_verify_options
 *
 *	Check each of the options to make sure they are valid for creating
 *	a new region.
 */
int aix_create_region_verify_options( aix_region_options_t * region_opts )
{
	int rc = 0;

	LOG_ENTRY();

	// Make sure the specified name is not in use.
	rc = aix_check_lv_name(region_opts->name);
	if (rc) {
		MESSAGE("Cannot use \"%s\" as name for new region.\n", region_opts->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Make sure the number of LPs is in the valid range. This does
	// not check for available PPs to allocate - that comes later.
	if ( region_opts->num_lps <= 0 ) {
		MESSAGE("Invalid number of LPs (%ld) for new region (%s).\n", region_opts->num_lps, region_opts->name);
		MESSAGE("Number of LPs must be a positive, non-zero number.\n");
		region_opts->num_lps = 1;
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}
	if ( region_opts->num_lps > AIX_MAX_LPS ) {
		MESSAGE("Invalid number of LPs (%ld) for new region (%s).\n", region_opts->num_lps, region_opts->name);
		MESSAGE("Number of LPs cannot exceed %d.\n", AIX_MAX_LPS);
		region_opts->num_lps = AIX_MAX_LPS;
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/* Function: aix_create_region_allocate_options
 *
 *	This is a helper function for aix_init_options. It sets up the initial
 *	information in the option descriptor for a Create task.
 */
int aix_create_region_allocate_option_descriptor( option_desc_array_t * od )
{
	LOG_ENTRY();

	od->count = AIX_OPTION_REGION_NUM_LPS_INDEX + 1;

	// Option 0 is Name
	SET_STRING(od->option[AIX_OPTION_REGION_NAME_INDEX].name, AIX_OPTION_REGION_NAME_STR);
	SET_STRING(od->option[AIX_OPTION_REGION_NAME_INDEX].title, "Name for new region");
	SET_STRING(od->option[AIX_OPTION_REGION_NAME_INDEX].tip, "Must be less than 64 characters");
	od->option[AIX_OPTION_REGION_NAME_INDEX].type           = EVMS_Type_String;
	od->option[AIX_OPTION_REGION_NAME_INDEX].unit           = EVMS_Unit_None;
	od->option[AIX_OPTION_REGION_NAME_INDEX].min_len    = AIX_MIN_NAMESIZ;
	od->option[AIX_OPTION_REGION_NAME_INDEX].max_len        = AIX_NAMESIZ;
	od->option[AIX_OPTION_REGION_NAME_INDEX].flags          = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
	od->option[AIX_OPTION_REGION_NAME_INDEX].constraint_type = EVMS_Collection_None;

	// Option 1 is Number of LPs
	SET_STRING(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].name, AIX_OPTION_REGION_NUM_LPS_STR);
	SET_STRING(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].title, "Number of LPs");
	SET_STRING(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].tip, "Number of Logical Partitions to allocate to the new region");
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].type        = EVMS_Type_Unsigned_Int32;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].unit        = EVMS_Unit_None;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].min_len     = 0;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].max_len     = 0;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].flags       = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;

	LOG_EXIT_INT(0);
	return 0;
}

/* Function: aix_split_region_allocate_options
 *
 *	This is a helper function for aix_init_options. It sets up the initial
 *	information in the option descriptor for a Create task.
 */
int aix_split_region_allocate_option_descriptor( option_desc_array_t * od )
{
	LOG_ENTRY();

	od->count = AIX_OPTION_REGION_NUM_LPS_INDEX;

	// Option 0 is Name
	SET_STRING(od->option[AIX_OPTION_REGION_NAME_INDEX].name, AIX_OPTION_REGION_NAME_STR);
	SET_STRING(od->option[AIX_OPTION_REGION_NAME_INDEX].title, "Name for new region");
	SET_STRING(od->option[AIX_OPTION_REGION_NAME_INDEX].tip, "Must be less than 64 characters");
	od->option[AIX_OPTION_REGION_NAME_INDEX].type           = EVMS_Type_String;
	od->option[AIX_OPTION_REGION_NAME_INDEX].unit           = EVMS_Unit_None;
	od->option[AIX_OPTION_REGION_NAME_INDEX].min_len    = AIX_MIN_NAMESIZ;
	od->option[AIX_OPTION_REGION_NAME_INDEX].max_len        = AIX_NAMESIZ;
	od->option[AIX_OPTION_REGION_NAME_INDEX].flags          = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
	od->option[AIX_OPTION_REGION_NAME_INDEX].constraint_type = EVMS_Collection_None;


	LOG_EXIT_INT(0);
	return 0;
}


/* Function: aix_create_region_get_acceptable
 *
 *	This is a helper function for aix_init_task. It determines which
 *	objects in EVMS are appropriate for a Create task.
 */
int aix_create_region_get_acceptable( dlist_t acceptable_objects )
{
	storage_container_t     * container;
	aix_container_data_t    * c_data;
	int                     rc;

	LOG_ENTRY();

	// Find all AIX freespace regions.
	DLIST_FOR_EACH(container, aix_container_list, rc) {
		c_data = container->private_data;
		if ( c_data->freespace->size > 0 ) {
			aix_add_object_to_list(c_data->freespace, acceptable_objects, AppendToList);
		}
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* Function: aix_create_region_set_objects
 *
 *	Examine the selected_objects list in the task. For a Create task, the
 *	first/only item on the list should be a freespace region. After the
 *	freespace region is selected, the initial values can be updated in the
 *	option descriptor.
 */
int aix_create_region_set_objects(      task_context_t  * context,
					dlist_t         declined_objects,
					task_effect_t   * effect )
{
	option_desc_array_t     * od = context->option_descriptors;
	storage_object_t        * region;
	aix_region_data_t       * r_data;
	unsigned long           max_lps;

	LOG_ENTRY();

	// Get the first item from the list.
	region = aix_get_freespace_region(context->selected_objects);
	if ( ! region ) {
		MESSAGE("Error getting freespace region.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	r_data = region->private_data;
	max_lps = (r_data->lv->num_lps <= AIX_MAX_LPS) ? r_data->lv->num_lps : AIX_MAX_LPS;

	// Make sure this freespace region has space available.
	if ( region->size <= 0 ) {
		MESSAGE("Selected freespace has no space available.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Update the constraint list for Number of LPs.
	SET_RANGE32(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].constraint.range, 1, max_lps, 1);
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].constraint_type     = EVMS_Collection_Range;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].value.ui32          = max_lps;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].flags               &= ~EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;

	*effect |= EVMS_Effect_Reload_Options;
	LOG_DEBUG("Selected freespace region %s.\n", region->name);
	LOG_EXIT_INT(0);
	return 0;
}


/* aix_create_region_set_option
 *
 *	This is a helper function for aix_set_option. It sets one particular
 *	option in the option descriptor for a Create task, and updates other
 *	option descriptor information accordingly.
 */
int aix_create_region_set_option(       task_context_t  * context,
					u_int32_t       index,
					value_t         * value,
					task_effect_t   * effect )
{
	option_desc_array_t     * od = context->option_descriptors;
	storage_object_t        * region;
	aix_region_data_t       * r_data;
	unsigned long           max_lps;
	int                     rc = 0;

	LOG_ENTRY();
	LOG_EXTRA("Setting option %d.\n", index);

	// Get the first item from the list.
	region = aix_get_freespace_region(context->selected_objects);
	if ( ! region ) {
		MESSAGE("Error getting freespace region.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}
	r_data = region->private_data;

	switch(index) {
	
	case AIX_OPTION_REGION_NAME_INDEX:
		// Make sure the name is not in use.
		rc = aix_check_lv_name(value->s);
		if (rc) {
			MESSAGE("Cannot use \"%s\" as name for new region.\n", value->s);
		}
		else {
			// Copy the name to the option descriptor array.
			SET_STRING(od->option[index].value.s, value->s);
			od->option[index].flags &= ~EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
			LOG_DEBUG("Set name option: %s\n", value->s);
		}
		break;

	case AIX_OPTION_REGION_NUM_LPS_INDEX:
		// Make sure a valid number of LPs have been selected.
		max_lps = (r_data->lv->num_lps <= AIX_MAX_LPS) ? r_data->lv->num_lps : AIX_MAX_LPS;
		if ( value->ui32 > max_lps ) {
			MESSAGE("%d LPs chosen. Only %ld available.\n", value->ui32, max_lps);
			value->ui32 = max_lps;
			*effect |= EVMS_Effect_Inexact;
		}

		od->option[index].value.ui32 = value->ui32;
		od->option[index].flags &= ~EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
		LOG_DEBUG("Set Number of LPs: %d\n", value->ui32);
		break;

	default:
		LOG_ERROR("Invalid option index: %d\n", index);
		rc = EINVAL;
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/* aix_split_region_set_option
 *
 *	This is a helper function for aix_set_option. It sets one particular
 *	option in the option descriptor for a Split task, and updates other
 *	option descriptor information accordingly.
 */
int aix_split_region_set_option(        task_context_t  * context,
					u_int32_t       index,
					value_t         * value,
					task_effect_t   * effect )
{
	option_desc_array_t     * od = context->option_descriptors;
	int                     rc = 0;

	LOG_ENTRY();

	switch(index) {
	
	case AIX_OPTION_REGION_NAME_INDEX:
		// Make sure the name is not in use.
		rc = aix_check_lv_name(value->s);
		if (rc) {
			LOG_DEBUG("Cannot use \"%s\" as name for new region.\n", value->s);
		}
		else {
			// Copy the name to the option descriptor array.
			SET_STRING(od->option[index].value.s, value->s);
			od->option[index].flags &= ~EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
			LOG_DEBUG("Set name option: %s\n", value->s);
		}
		break;

	default:
		LOG_ERROR("Invalid option index: %d\n", index);
		rc = EINVAL;
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/********** Region Expand Options **********/



/* aix_expand_region_parse_options
 *
 *	Examine the option array for region creation options. Current valid
 *	options are:
 *	NAME
 *	NUM_LPS
 */
int aix_expand_region_parse_options(option_array_t          * options,
				    storage_object_t        * region,
				    aix_region_options_t    * region_opts )
{
	unsigned int    i;
	int             rc;

	LOG_ENTRY();

	memset(region_opts, 0, sizeof(aix_region_options_t));

	for ( i = 0; i < options->count; i++ ) {
		// If only the name-based index is specified, get the number.
		if ( ! options->option[i].is_number_based ) {
			if ( ! strcmp(options->option[i].name, AIX_OPTION_REGION_NAME_STR) ) {
				options->option[i].number = AIX_OPTION_REGION_NAME_INDEX;
			}
			else if ( ! strcmp(options->option[i].name, AIX_OPTION_REGION_NUM_LPS_STR) ) {
				options->option[i].number = AIX_OPTION_REGION_NUM_LPS_INDEX;
			}
			else {
				continue;
			}
		}

		LOG_EXTRA("Parsing option %d\n", options->option[i].number);

		// Use the number-based index to record the option.
		switch ( options->option[i].number ) {
		case AIX_OPTION_REGION_NAME_INDEX:
			strncpy(region_opts->name, options->option[i].value.s, AIX_NAMESIZ-1);
			LOG_DEBUG("Setting name: %s\n", region_opts->name);
			break;
		case AIX_OPTION_REGION_NUM_LPS_INDEX:
			region_opts->num_lps = options->option[i].value.ui32;
			LOG_DEBUG("Setting number of LPs: %ld\n", region_opts->num_lps);
			break;
		default:
			break;
		}
	}

	rc = aix_expand_region_verify_options(region_opts, region);

	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_expand_region_verify_options
 *
 *	Check each of the options to make sure they are valid for creating
 *	a new region.
 */
int aix_expand_region_verify_options( aix_region_options_t * region_opts,
				      storage_object_t     * region)
{
	int rc = 0;

	LOG_ENTRY();

	// Make sure the number of LPs is in the valid range. This does
	// not check for available PPs to allocate - that comes later.
	if ( region_opts->num_lps <= 0 ) {
		MESSAGE("Invalid number of LPs (%ld) for new region (%s).\n", region_opts->num_lps, region_opts->name);
		MESSAGE("Number of LPs must be a positive, non-zero number.\n");
		region_opts->num_lps = 1;
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}
	if ( region_opts->num_lps > AIX_MAX_LPS ) {
		MESSAGE("Invalid number of LPs (%ld) for new region (%s).\n", region_opts->num_lps, region_opts->name);
		MESSAGE("Number of LPs cannot exceed %d.\n", AIX_MAX_LPS);
		region_opts->num_lps = AIX_MAX_LPS;
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}


	LOG_EXIT_INT(rc);
	return rc;
}
/********** Region Shrink Options **********/



/* aix_shrink_region_parse_options
 *
 *	Examine the option array for region creation options. Current valid
 *	options are:
 *
 *	NUM_LPS
 */
int aix_shrink_region_parse_options(option_array_t          * options,
				    storage_object_t        * region,
				    aix_region_options_t    * region_opts )
{
	int i, rc;

	LOG_ENTRY();

	memset(region_opts, 0, sizeof(aix_region_options_t));

	for ( i = 0; i < options->count; i++ ) {
	
		if ( ! options->option[i].is_number_based ) {

 			if ( ! strcmp(options->option[i].name, AIX_OPTION_REGION_NUM_LPS_STR) ) {

				region_opts->num_lps = options->option[i].value.ui32;
				
			}			

		}

	}

	LOG_DEBUG("Setting number of LPs: %ld\n", region_opts->num_lps);

	rc = aix_shrink_region_verify_options(region_opts, region);

	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_shrink_region_verify_options
 *
 *	Check each of the options to make sure they are valid for creating
 *	a new region.
 */
int aix_shrink_region_verify_options( aix_region_options_t * region_opts,
				      storage_object_t     * region)
{
	int rc = EINVAL;
	aix_region_data_t * r_data = region->private_data;

	LOG_ENTRY();

	if ( region_opts->num_lps > 0 &&
	     region_opts->num_lps < r_data->lv->num_lps) {
		rc = 0;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/* Function: aix_expand_container_allocate_options
 *
 *	This is a helper function for aix_init_options. It sets up the initial
 *	information in the option descriptor for a Expand_Container task.
 */
int aix_expand_container_allocate_option_descriptor( option_desc_array_t * od )
{
	LOG_ENTRY();

	od->count = 0;

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_expand_container_parse_options
 *
 *	Examine the option array for container expantion options. Current valid
 *	options are:
 *	PP_SIZE
 *
 *	The "pp_size" argument to this function is in vsectors, NOT in
 *	exponent form.
 */
int aix_expand_container_parse_options( option_array_t  * options,
					unsigned long   * pp_size )
{
	unsigned int    i;
	int             rc = 0;

	LOG_ENTRY();

	// Default PP size is 4MB.
	//*pp_size = AIX_DEFAULT_PP_SIZE;

	for ( i = 0; i < options->count; i++ ) {
		// If only the name-based index is specified, get the number.
		if ( ! options->option[i].is_number_based ) {
			if ( ! strcmp(options->option[i].name, AIX_OPTION_CONTAINER_PP_SIZE_STR) ) {
				options->option[i].number = AIX_OPTION_CONTAINER_PP_SIZE_INDEX;
			}
			else {
				continue;
			}
		}

		LOG_EXTRA("Parsing option %d\n", options->option[i].number);

		// Use the number-based index to record the option.
		switch ( options->option[i].number ) {
		case AIX_OPTION_CONTAINER_PP_SIZE_INDEX:
			*pp_size = options->option[i].value.ui32;
			LOG_DEBUG("Setting PP size: %ld sectors\n", *pp_size);
			break;
		default:
			break;
		}
	}

	// We've already checked this once before we got here
	//
	//rc = aix_create_container_verify_options(pp_size);

	LOG_EXIT_INT(rc);
	return rc;
}
/* Function: aix_expand_container_get_acceptable
 *
 *	This is a helper function for aix_init_task. It determines which
 *	objects in EVMS are appropriate for a Exapnd_Container task.
 */
int aix_expand_container_get_acceptable( storage_container_t * container,
					 dlist_t acceptable_objects )
{
	storage_object_t         * object;
	dlist_t                 object_list;
	sector_count_t          num_sectors;
	aix_container_data_t * c_data;
	int                     rc;

	LOG_ENTRY();

	// Find all top-level disks, segments, and regions in the system. This
	// will also retrieve AIX regions, so let the user beware. :)
	rc = EngFncs->get_object_list(DISK|SEGMENT|REGION, DATA_TYPE, 0, NULL, TOPMOST, &object_list);
	if (rc) {
		LOG_EXIT_INT(rc);
		return rc;
	}

	c_data = container->private_data;

	// Don't add objects that are too small or too big to use with AIX.

	DLIST_FOR_EACH(object, object_list, rc) {
		num_sectors = AVAILABLE_PP_SECTORS(object);

		LOG_DEBUG("Checking Container: [%s] \n", container->name);

		if ( num_sectors > AIX_DEFAULT_PP_SIZE &&
		     num_sectors / AIX_MAX_PP_SIZE <= AIX_MAX_PPS_PER_PV &&
		     num_sectors > PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size)) {

			LOG_DEBUG("Adding Container: [%s] Segment log2:%ld Container log2:%d\n", container->name,
				  VSECTORS_TO_PP_SIZE(num_sectors), c_data->vgda->vg_head->pp_size);
			aix_add_object_to_list(object, acceptable_objects, AppendToList);
		}
	}

	DestroyList(&object_list, FALSE);

	LOG_EXIT_INT(0);
	return 0;
}

/* Function: aix_expand_region_allocate_options
 *
 *	This is a helper function for aix_init_options. It sets up the initial
 *	information in the option descriptor for an Expand task.
 */
int aix_expand_region_allocate_option_descriptor( option_desc_array_t * od )
{
	LOG_ENTRY();

	od->count = AIX_OPTION_REGION_NUM_LPS_INDEX + 1;

	// Option 0 is Size
	SET_STRING(od->option[AIX_OPTION_REGION_EXPAND_SIZE_INDEX].name, AIX_OPTION_REGION_EXPAND_SIZE_STR);
	SET_STRING(od->option[AIX_OPTION_REGION_EXPAND_SIZE_INDEX].title, "Additional Size");
	SET_STRING(od->option[AIX_OPTION_REGION_EXPAND_SIZE_INDEX].tip, "Amount of space to add to the selected AIX region");
	od->option[AIX_OPTION_REGION_EXPAND_SIZE_INDEX].type            = EVMS_Type_Unsigned_Int32;
	od->option[AIX_OPTION_REGION_EXPAND_SIZE_INDEX].unit            = EVMS_Unit_None;
	od->option[AIX_OPTION_REGION_EXPAND_SIZE_INDEX].min_len     = 0;
	od->option[AIX_OPTION_REGION_EXPAND_SIZE_INDEX].max_len     = 0;
	od->option[AIX_OPTION_REGION_EXPAND_SIZE_INDEX].flags           = EVMS_OPTION_FLAGS_INACTIVE;
	od->option[AIX_OPTION_REGION_EXPAND_SIZE_INDEX].constraint_type = EVMS_Collection_None;

	// Option 1 is Number of LPs
	SET_STRING(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].name, AIX_OPTION_REGION_NUM_LPS_STR);
	SET_STRING(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].title, "Number of LPs");
	SET_STRING(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].tip, "Number of Logical Paritions to allocate to the new region");
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].type        = EVMS_Type_Unsigned_Int32;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].unit        = EVMS_Unit_None;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].min_len     = 0;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].max_len     = 0;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].flags       = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;

	LOG_EXIT_INT(0);
	return 0;
}



/* Function: aix_expand_region_get_acceptable
 *
 *	This is a helper function for aix_init_task. It determines which
 *	objects in EVMS are appropriate for an Expand task.
 */
int aix_expand_region_get_acceptable( storage_object_t    * region,
				      dlist_t               acceptable_objects )
{
	storage_container_t     * container;
	aix_container_data_t    * c_data;
	aix_container_data_t    * region_c_data = region->producing_container->private_data;
	int                     rc;

	LOG_ENTRY();

	// Find all AIX freespace containers/regions. We can only expand within our own region
	DLIST_FOR_EACH(container, aix_container_list, rc) {
		c_data = container->private_data;
		if ( c_data->freespace->size > 0 && (COMPARE_UNIQUE_IDS(region_c_data->vgda->vg_head->vg_id, c_data->vgda->vg_head->vg_id))){

			aix_add_object_to_list(c_data->freespace, acceptable_objects, AppendToList);
		}
	}

	LOG_EXIT_INT(0);
	return 0;
}

/* Function: aix_shrink_region_allocate_options
 *
 *	This is a helper function for aix_init_options. It sets up the initial
 *	information in the option descriptor for an Shrink task.
 */
int aix_shrink_region_allocate_option_descriptor( option_desc_array_t * od )
{
	LOG_ENTRY();

	od->count = AIX_OPTION_REGION_NUM_LPS_INDEX;


	// Option 0 is Number of LPs
	SET_STRING(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].name, AIX_OPTION_REGION_NUM_LPS_STR);
	SET_STRING(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].title, "Number of LPs");
	SET_STRING(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].tip, "Number of Logical Paritions to remove from the region");
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].type      = EVMS_Type_Unsigned_Int32;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].unit      = EVMS_Unit_None;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].min_len   = 0;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].max_len   = 0;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].flags     = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;

	LOG_EXIT_INT(0);
	return 0;
}



/* Function: aix_expand_region_get_acceptable
 *
 *	This is a helper function for aix_init_task. It determines which
 *	objects in EVMS are appropriate for an Shrink task.
 */
int aix_shrink_region_get_acceptable( storage_object_t    * region,
				      dlist_t               acceptable_objects )
{

	LOG_ENTRY();

	// Find all AIX freespace regions.
	if ( region ) {
		aix_add_object_to_list(region, acceptable_objects, AppendToList);
	}

	LOG_EXIT_INT(0);
	return 0;
}

/* Function: aix_expand_container_set_objects
 *
 *	Examine the selected_objects list in the task, and verify that each
 *	object is valid for inclusion in a new container. If any are not valid,
 *	remove them from the selected_objects and add them to the
 *	declined_objects. In create_container, no options will change based
 *	on which objects are selected.
 */
int aix_expand_container_set_objects(   task_context_t  * context,
					dlist_t         declined_objects,
					task_effect_t   * effect )
{
	aix_container_data_t * c_data = context->container->private_data;
	storage_object_t         * object;
	unsigned long          pp_size = PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size);
	int                     rc;

	LOG_ENTRY();

	// Check that all of the objects are valid for use in a new container.
	DLIST_FOR_EACH(object, context->selected_objects, rc) {

		rc = aix_check_object_for_container_inclusion(object);
		if (rc) {
// Add this object to declined objects.
			LOG_ERROR("One or more objects are invalid for container creation.\n");
			return(EINVAL);
		}

		// If any of the segments are too small for the default PE
		// size, try to reset the PE size.
		rc = aix_check_object_for_pp_size(object, &pp_size);
		if (rc == EINVAL) {
// Add this object to declined objects.
			LOG_ERROR("One or more objects are invalid for container creation.\n");
			return(EINVAL);
		}
		else
			if (rc) {
			LOG_ERROR("One or more objects are invalid for container expansion\n");
			return(EINVAL);
		}
	}

	LOG_EXIT_INT(0);
	return 0;

}
/* Function: aix_expand_region_set_objects
 *
 *	Examine the selected_objects list in the task. For an Exapnd task, the
 *	first/only item on the list should be a freespace region. After the
 *	freespace region is selected, the initial values can be updated in the
 *	option descriptor.
 */
int aix_shrink_region_set_objects(      task_context_t  * context,
					dlist_t         declined_objects,
					task_effect_t   * effect )
{
	option_desc_array_t         * od = context->option_descriptors;
	storage_object_t            * region;
	aix_region_data_t           * r_data;
	unsigned long             max_lps;

	LOG_ENTRY();



	// Get the first item from the list.
	region = context->object;
	if ( ! region ) {
		MESSAGE("Error getting freespace region.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}


	r_data = region->private_data;
	max_lps = r_data->lv->num_lps - 1;

	// Make sure this freespace region has space available.
	if ( region->size <= 0 ) {
		MESSAGE("Selected freespace has no space available.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Update the constraint list for Number of LPs.
	SET_RANGE32(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].constraint.range, 1, max_lps, 1);
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].constraint_type   = EVMS_Collection_Range;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].value.ui32                = max_lps;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX-1].flags             &= ~EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;

	*effect |= EVMS_Effect_Reload_Options;
	LOG_DEBUG("Selected freespace region [%s] max_lps = %lu\n", region->name, max_lps);
	LOG_EXIT_INT(0);
	return 0;
}

/* Function: aix_expand_region_set_objects
 *
 *	Examine the selected_objects list in the task. For an Exapnd task, the
 *	first/only item on the list should be a freespace region. After the
 *	freespace region is selected, the initial values can be updated in the
 *	option descriptor.
 */
int aix_expand_region_set_objects(      task_context_t  * context,
					dlist_t         declined_objects,
					task_effect_t   * effect )
{
	option_desc_array_t         * od = context->option_descriptors;
	storage_object_t            * region;
	aix_region_data_t           * r_data;
	unsigned long             max_lps;

	LOG_ENTRY();

	// Get the first item from the list.
	region = aix_get_freespace_region(context->selected_objects);
	if ( ! region ) {
		MESSAGE("Error getting freespace region.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}


	r_data = region->private_data;
	max_lps = (r_data->lv->num_lps <= AIX_MAX_LPS) ? r_data->lv->num_lps : AIX_MAX_LPS;

	// Make sure this freespace region has space available.
	if ( region->size <= 0 ) {
		MESSAGE("Selected freespace has no space available.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Update the constraint list for Number of LPs.
	SET_RANGE32(od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].constraint.range, 1, max_lps, 1);
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].constraint_type     = EVMS_Collection_Range;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].value.ui32          = max_lps;
	od->option[AIX_OPTION_REGION_NUM_LPS_INDEX].flags               &= ~EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;

	*effect |= EVMS_Effect_Reload_Options;
	LOG_DEBUG("Selected freespace region [%s] max_lps = %lu\n", region->name, max_lps);
	LOG_EXIT_INT(0);
	return 0;
}

/* aix_expand_region_set_option
 *
 *	This is a helper function for aix_set_option. It sets one particular
 *	option in the option descriptor for a Create task, and updates other
 *	option descriptor information accordingly.
 */
int aix_shrink_region_set_option(       task_context_t  * context,
					u_int32_t       index,
					value_t         * value,
					task_effect_t   * effect )
{
	option_desc_array_t * od = context->option_descriptors;
	int         rc = 0;

	LOG_ENTRY();
	LOG_DEBUG("Setting option %d.\n", index);


	od->option[index].value.ui32 = value->ui32;
	od->option[index].flags &= ~EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
	LOG_DEBUG("Set Number of LPs: %d\n", value->ui32);

	LOG_EXIT_INT(rc);
	return rc;
}
/* aix_expand_region_set_option
 *
 *	This is a helper function for aix_set_option. It sets one particular
 *	option in the option descriptor for a Create task, and updates other
 *	option descriptor information accordingly.
 */
int aix_expand_region_set_option(       task_context_t  * context,
					u_int32_t       index,
					value_t         * value,
					task_effect_t   * effect )
{
	option_desc_array_t     * od = context->option_descriptors;
	storage_object_t        * region;
	aix_region_data_t       * r_data;
	unsigned long           max_lps;
	int                     rc = 0;

	LOG_ENTRY();
	LOG_EXTRA("Setting option %d.\n", index);

	// Get the first item from the list.
	region = aix_get_freespace_region(context->selected_objects);
	if ( ! region ) {
		MESSAGE("Error getting freespace region.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}
	r_data = region->private_data;

	switch(index) {
	
	case AIX_OPTION_REGION_EXPAND_SIZE_INDEX:

		break;

	case AIX_OPTION_REGION_NUM_LPS_INDEX:
		// Make sure a valid number of LPs have been selected.
		max_lps = (r_data->lv->num_lps <= AIX_MAX_LPS) ? r_data->lv->num_lps : AIX_MAX_LPS;
		if ( value->ui32 > max_lps ) {
			MESSAGE("%d LPs chosen. Only %ld available.\n", value->ui32, max_lps);
			value->ui32 = max_lps;
			*effect |= EVMS_Effect_Inexact;
		}

		od->option[index].value.ui32 = value->ui32;
		od->option[index].flags &= ~EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
		LOG_DEBUG("Set Number of LPs: %d\n", value->ui32);
		break;

	default:
		LOG_ERROR("Invalid option index: %d\n", index);
		rc = EINVAL;
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}
/* Function: aix_expand_container_set_option
 *
 *	This is a helper function for aix_set_option. It sets one particular
 *	option in the option descriptor for a Create_Container task, and
 *	updates other option descriptor information accordingly.
 */
int aix_expand_container_set_option(    task_context_t  * context,
					u_int32_t       index,
					value_t         * value,
					task_effect_t   * effect )
{
	option_desc_array_t      * od = context->option_descriptors;
	aix_container_data_t * c_data = context->container->private_data;
	storage_object_t         * object;
	unsigned long          pp_size = PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size);
	int                     rc = 0;

	LOG_ENTRY();
	LOG_EXTRA("Setting option %d\n", index);

	switch (index) {
	
	case AIX_OPTION_CONTAINER_PP_SIZE_INDEX:
		//pp_size = value->ui32;

		rc = aix_check_pp_size(&pp_size);
		if (rc) {
			*effect |= EVMS_Effect_Inexact;
		}

		// Make sure every object in the "selected" list is large
		// enough for the specified pp_size.
		DLIST_FOR_EACH(object, context->selected_objects, rc) {
			rc = aix_check_object_for_pp_size(object, &pp_size);
			if (rc) {
				LOG_ERROR("One or more objects too small for PE size %d.\n", value->ui32);
				rc = ENOSPC;
				break;
			}
		}

		if ( rc != ENOSPC ) {
			od->option[index].value.ui32 = pp_size;
			od->option[index].flags &= ~EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
			rc = 0;
		}
		break;

	default:
		rc = EINVAL;
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/* Function: aix_list_acceptable_mirrors
 *
 *	This is a helper function for aix_add_mirror.
 *	
 *	
 */
int aix_list_acceptable_mirrors(storage_object_t *master_object,
				dlist_t          acceptable_objects)



{

	aix_region_data_t        *r_data = master_object->private_data;
	storage_object_t         *object;
	Insertion_Modes           InsertMode;
	int rc         = 0;
	int AllOnOnePV = TRUE;

	LOG_ENTRY();

	aix_check_physical_layout(master_object, &AllOnOnePV);

	DLIST_FOR_EACH(object, master_object->producing_container->objects_consumed, rc) {


		if (!AllOnOnePV) {
			InsertMode = InsertBefore;
		} else if (AllOnOnePV && r_data->lp_map[0].object == object) {
			InsertMode =  AppendToList;
		} else {
			InsertMode = InsertBefore;
		}

		if (r_data->lv->num_lps <= aix_get_free_pps_for_object(object, r_data)) {
			rc = aix_add_object_to_list(object, acceptable_objects, InsertMode);
		}
	}

	LOG_EXIT_INT(0);
	return 0;
}
int aix_get_available_objects(storage_object_t *region, int * count)
{

	aix_region_data_t   * r_data = region->private_data;
	storage_object_t    * object;
	int rc;

	LOG_ENTRY();

	*count = 0;

	DLIST_FOR_EACH(object, region->producing_container->objects_consumed, rc){
		if (r_data->lv->num_lps <= aix_get_free_pps_for_object(object, r_data)) {
			*count = *count+1;
		}
	}

	if (rc == DLIST_END_OF_LIST) {
		rc = DLIST_SUCCESS;
	}

	LOG_EXIT_INT(rc);
	return rc;

}
