/*
 *
 *   (C) Copyright IBM Corp. 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: libdrivelink.so
 *
 *   File: dl_options.c
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <plugin.h>

#include "dl_common.h"



/*
 *   Called to build a reverse ordered child_objects dlist from a list
 *   of child objects.
 *
 *   Returns DLIST_SUCCESS if successful and the child_objects
 *   dlist should contain all the objects from the drive link
 *   in the proper order.
 *
 *   Returns
 */
static int dl_build_reverse_ordered_child_object_list( storage_object_t *drivelink, dlist_t list )
{
	void  			 *handle=NULL;
	int   		      	  i, rc=EINVAL,dlist_rc;
	dlist_t  		  reverse_ordered_list;
	drivelink_private_data_t  *pdata;
	uint                      starting_count=0;
	uint                      ending_count=0;
	storage_object_t         *obj;

	LOG_ENTRY();

	REQUIRE(drivelink != NULL);
	REQUIRE(drivelink->private_data != NULL);
	REQUIRE(list != NULL);
	
	pdata = (drivelink_private_data_t  *) drivelink->private_data;

        GetListSize(list, &starting_count);

        reverse_ordered_list = CreateList();

	REQUIRE(reverse_ordered_list != NULL);

        for(i=pdata->drive_link_count-1; i>=0; i-- ) {

		if ( pdata->drive_link[i].object != NULL ) {

			DLIST_FOR_EACH(obj, list, dlist_rc) {
		
				if (obj == pdata->drive_link[i].object) {

					rc = ExclusiveInsertObject ( reverse_ordered_list,
								     obj,
								     obj->object_type,
								     NULL,
								     AppendToList,
								     TRUE,
								     &handle );
					if (rc == 0) {
						++ending_count;
					}

					break;
				}
                            
                        }
		}

        }
 
        // if I finished with a reversed ordered list that contains
        // the same number of objects that I started with ... then
        // it means that the drive link ordering table matches the
        // input list exactly.
        //
        // So, now copy the reversed ordered list back to the callers
        // list so we can return the reordered drive link objects.
        if ( rc == DLIST_SUCCESS ) {

		if (starting_count == ending_count) {

			// first delete all objects in the caller list
			rc = DeleteAllItems ( list, FALSE );

			// then copy reordered objects to callers list
			if (rc == DLIST_SUCCESS) {
				rc = CopyList(list, reverse_ordered_list, AppendToList);
			}

		}
		else {
			rc = EPERM;
		}

        }

	// dont forget to toss working dlist
	if (reverse_ordered_list != NULL) DestroyList(&reverse_ordered_list, FALSE);

	LOG_EXIT_INT(rc);
	return rc;
}





/*
 *  Returns count of options for specified task
 */
int dl_get_option_count(task_context_t * task)
{
	int count;

	LOG_ENTRY();

	switch (task->action) {

		case EVMS_Task_Create:
		case EVMS_Task_Add_Feature:
			count = DL_CREATE_OPTION_COUNT;
			break;

		case EVMS_Task_Expand:
			count = DL_EXPAND_OPTION_COUNT;
			break;

		case EVMS_Task_Shrink:
			count = DL_SHRINK_OPTION_COUNT;
			break;

		case EVMS_Replace_Missing_DL_Child:
			count = DL_REPLACE_MISSING_CHILD_OPTION_COUNT;
			break;

		default:
			count = 0;
			break;
		}

	LOG_EXIT_INT(count);
	return count;
}

/*
 *  Called by init_task() to allocate the REPLACE option descriptor array and
 *  setup the individual option descriptors.  It does not initialize the
 *  descriptors.
 */
static int allocate_replace_option_descriptors( task_context_t * context )
{
	int rc;	

	LOG_ENTRY();
	rc=0;
	LOG_EXIT_INT(rc);
	return rc;
}


/*
 *  Called to initialize the REPLACE options in the task context.
 */
static int initialize_replace_option_descriptors( storage_object_t  * missing_child, 
						  task_context_t    * context )
{
	int rc;

	LOG_ENTRY();
	rc=0;
	LOG_EXIT_INT(rc);
	return rc;
}


/*
 *  Function: set replace object
 *
 *  Called by set_object() to validate the task selected 
 *  object.
 */
static int set_replace_object( task_context_t * context,
			       dlist_t          declined_objects,
			       task_effect_t  * effect )
{
    int  rc;
    storage_object_t  *new_child;
    void              *handle;
    boolean            found_good_object = FALSE;
    declined_object_t *declined_object=NULL;


    LOG_ENTRY();

    rc = GoToStartOfList( context->selected_objects );

    while (rc == DLIST_SUCCESS) {

        new_child = get_object_from_list( context->selected_objects );
        if (new_child) {

            if ( found_good_object == FALSE ) {

                    found_good_object = TRUE;

                    rc = initialize_replace_option_descriptors(new_child, context);

                    if (rc == DLIST_SUCCESS) {
                        *effect |=  EVMS_Effect_Reload_Options;
                        rc = NextItem(context->selected_objects);
                    }

            }
            else {

                declined_object = dl_engine_fncs->engine_alloc( sizeof(declined_object_t));

                if (declined_object) {

                    declined_object->object = new_child;
                    declined_object->reason = EINVAL;

                    rc = InsertObject ( declined_objects,
                                        declined_object,
                                        DECLINED_OBJECT_TAG,
                                        NULL,
                                        AppendToList,
                                        TRUE,
                                        &handle );

                    if (rc == DLIST_SUCCESS) {
                        *effect |=  EVMS_Effect_Reload_Objects;
                        rc = NextItem(context->selected_objects);
                    }
                    else {
                        dl_engine_fncs->engine_free(declined_object);
                    }

                }
                else {
                    LOG_ERROR("unable to malloc a declined object struct\n");
                    rc = ENOMEM;
                }

            }


        }
        else {
            rc = DLIST_END_OF_LIST;
        }

    };


    // throw away errors like ... DLIST_END_OF_LIST ... which are not real errors
    if ( ( rc == DLIST_EMPTY )||( rc == DLIST_END_OF_LIST )) {
        rc = DLIST_SUCCESS;
    }


    LOG_EXIT_INT(rc);
    return rc;
}

/*
 *  Function: prune invalid replace targets
 *
 *  Called by get_acceptable_replace_objects() to prune
 *  any object that cannot be used as a move target for
 *  the specified missing child object.
 */
static boolean prune_invalid_replace_targets( ADDRESS   Object,
					      TAG       ObjectTag,
					      ADDRESS   ObjectHandle,
					      ADDRESS   Parameters,
					      boolean  *FreeMemory,
					      uint     *Error )
{
 	storage_object_t  *new_child = (storage_object_t *)Object;
	storage_object_t  *missing_child = (DISKSEG *)Parameters;
	storage_object_t  *drivelink;
	int rc=0;

	*FreeMemory = FALSE;           // tells dlist not to free any memory
	*Error      = DLIST_SUCCESS;   // tells dlist we were successful
		
	drivelink = dl_get_parent(missing_child);

	if (drivelink != new_child) {
		rc = dl_validate_missing_child_replace_target(missing_child, new_child);
		if (rc == 0) {
			return FALSE; // tells dlist NOT to prune object from list
		}
	}
	
	return TRUE;           // tells dlist to prune this object from acceptable DISK list
}

/*
 *  Function: get acceptable replace objects
 *
 *  Called when setting up a replace task to go out and find all the
 *  objects that we could replace a missing child with. 
 */
static int get_acceptable_replace_objects( task_context_t * context )
{
	int   rc = EINVAL;
	uint  count=0;	

	LOG_ENTRY();

	if ( (context) &&
             (context->object) &&
             (context->acceptable_objects) ) {

		GetListSize(context->acceptable_objects, &count);		
		if (count == 0) {

			rc = dl_engine_fncs->get_object_list( 0,
							      DATA_TYPE,
							      NULL,
							      NULL,
							      TOPMOST,
							      &context->acceptable_objects );

			if (rc == DLIST_SUCCESS) {
				PruneList( context->acceptable_objects, 
					   prune_invalid_replace_targets,
					   context->object );
			}			

		}
		else {
			LOG_ERROR("error, context already has acceptable objects\n");
		}

	}

	LOG_EXIT_INT(rc);
	return rc;
}





static int dl_allocate_create_option_descriptors( task_context_t * context )
{
	int rc=0;


	LOG_ENTRY();

	REQUIRE(context != NULL);
   
        context->option_descriptors->count = DL_CREATE_OPTION_COUNT ;     

        context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].flags = EVMS_OPTION_FLAGS_INACTIVE ;
        context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].constraint.list = NULL;
        context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].constraint_type = EVMS_Collection_None;
        context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].help = NULL;
        SET_STRING( context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].name, DL_CREATE_LINKNAME_NAME );
        context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].min_len = 1;
        context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].max_len = EVMS_VOLUME_NAME_SIZE;
        SET_STRING( context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].tip, "The name to be given to the drivelink object.");
        SET_STRING( context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].title, "Name" );
        context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].type = EVMS_Type_String;
        context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].unit = EVMS_Unit_None;
        context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].value.s = dl_engine_fncs->engine_alloc(EVMS_VOLUME_NAME_SIZE+1);
        if (context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].value.s) {
		memset( context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].value.s,
			0, 
			EVMS_VOLUME_NAME_SIZE);
        }
        else {
		rc = ENOMEM;
		LOG_EXIT_INT(rc);
		return rc;
        }

	LOG_EXIT_INT(rc);
	return rc;
}


static int dl_initialize_create_option_descriptors(  task_context_t * context )
{
	int rc=0;

	LOG_ENTRY();

	REQUIRE(context != NULL);

        // there should be exactly DL_CREATE_OPTION_COUNT options
        if ( context->option_descriptors->count == DL_CREATE_OPTION_COUNT ) {
		context->option_descriptors->option[DL_CREATE_LINKNAME_INDEX].flags = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
        }
        else {
		LOG_ERROR("error, wrong number of option descriptors ... count= %d\n", context->option_descriptors->count);
		rc = EINVAL;
        }
   
	LOG_EXIT_INT(rc);
	return rc;
}


static int dl_allocate_expand_option_descriptors( task_context_t * context )
{
	int rc=0;

	LOG_ENTRY();

	REQUIRE(context != NULL);

        context->option_descriptors->count = 0;  // no options
        
	LOG_EXIT_INT(rc);
	return rc;
}


static int dl_allocate_shrink_option_descriptors( task_context_t * context )
{
	int rc=0;

	LOG_ENTRY();

	REQUIRE(context != NULL);

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 *  Get the user selected object from the create context and init the option descriptors
 *  as appropriate for the object.
 *
 *  Note ... if there is more than 1 object in the source object list we will simply
 *           take the first good object and decline the remainder.  That is the reason for
 *           the boolean flag found_good_object.
 */
static int dl_set_create_object( task_context_t * context,
                                 dlist_t          declined_objects,
                                 task_effect_t  * effect )
{
	int   rc;
	uint  linkcount=0;

	LOG_ENTRY();

	REQUIRE(context != NULL);
	REQUIRE(declined_objects != NULL);
	REQUIRE(effect != NULL);

	GetListSize(context->selected_objects, &linkcount );
    
	REQUIRE(linkcount > 0 && linkcount <= EVMS_DRIVELINK_MAX_ENTRIES);

        rc = dl_initialize_create_option_descriptors( context );

        if (rc == 0) {
                *effect |=  EVMS_Effect_Reload_Options;
        }
        else {
                rc = EINVAL;
        }

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Validate the objects in the source and target dlists in the context
 * record.  Remove from the selected objects lists any objects which are no
 * longer acceptable.  Return all acceptable objects in the parameter
 * dlists.  Also, for any object which is removed from the selected dlists,
 * or is otherwise not acceptable, create a declined_handle_t struct with
 * reason why not acceptable and add to the declined_objects dlist.
 */
static int dl_set_expand_object( task_context_t * context,
                                 dlist_t          declined_objects,
                                 task_effect_t  * effect )
{
	int                      rc=0,dlist_rc;
	storage_object_t        *obj=NULL;
	void                    *handle;
	declined_object_t       *declined_object=NULL;
	int                      selected_objects_count=0;
	int                      max_objects_allowed=0;
	drivelink_private_data_t *pdata;
	int                      declined_object_count=0;


	LOG_ENTRY();
	
	REQUIRE(context != NULL);
	REQUIRE(declined_objects != NULL);
	REQUIRE(effect != NULL);
	REQUIRE(dl_isa_drivelink(context->object) == TRUE);

        pdata = ( drivelink_private_data_t * )context->object->private_data;

        max_objects_allowed = EVMS_DRIVELINK_MAX_ENTRIES - pdata->drive_link_count;

	REQUIRE(max_objects_allowed > 0);

	DLIST_FOR_EACH(obj, context->selected_objects, dlist_rc){
        
		if ( max_objects_allowed > selected_objects_count ) {
			++selected_objects_count;
			*effect |=  EVMS_Effect_Reload_Options;			
		}
		else {

			LOG_ERROR("declining an object found in context->selected_objects, object name= %s\n", obj->name);

			++declined_object_count;

			declined_object = dl_engine_fncs->engine_alloc( sizeof(declined_object_t));

			if (declined_object) {

				declined_object->object = obj;
				declined_object->reason = rc; // REMINDER ... need to pickup reason codes the UI understands

				
				rc = InsertObject ( declined_objects,
						   declined_object,
						   DECLINED_OBJECT_TAG,
						   NULL,
						   AppendToList,
						   TRUE,
						   &handle );

				if (rc == DLIST_SUCCESS) {
					*effect |=  EVMS_Effect_Reload_Objects;
					rc = NextItem(context->selected_objects);
				}
				else {
					dl_engine_fncs->engine_free(declined_object);
					LOG_ERROR("error, unable to insert declined object in engine dlist\n");
				}

			}
			else {
				LOG_ERROR("unable to malloc a declined object struct\n");
				rc = ENOMEM;
			}

		}
	}

	if (declined_object_count) {
		*effect |= EVMS_Effect_Reload_Objects;
	}
	else if ( (rc==0) &&                        // didnt decline any of the objects and the result looks
                  (selected_objects_count > 0) ) {  // good so now call the engine to see if we can expand
                                                // by the specified amount.

		sector_count_t   expand_sectors = 0;
		sector_count_t   child_sectors  = 0;

		DLIST_FOR_EACH(obj, context->selected_objects, dlist_rc) {

			child_sectors = obj->size - (DRIVELINK_METADATA_SECTOR_COUNT*2) - (FEATURE_HEADER_SECTOR_COUNT*2);
			expand_sectors += child_sectors;

			rc = dl_engine_fncs->can_expand_by( obj, &expand_sectors );

			if (rc) {

				expand_sectors -= child_sectors;

				LOG_DEBUG("declining object, object name= %s ... engine wont allow expanding\n", obj->name);

				declined_object = dl_engine_fncs->engine_alloc( sizeof(declined_object_t));

				if (declined_object) {

					declined_object->object = obj;
					declined_object->reason = -1; // ENOCANEXPAND

					// remove object from selectable/acceptable object lists
					DeleteObject( context->selected_objects, obj );
					DeleteObject( context->acceptable_objects, obj );

					// place object in declined list
					rc = InsertObject ( declined_objects,
							    obj,
							    DECLINED_OBJECT_TAG,
							    NULL,
							    AppendToList,
							    TRUE,
							    &handle );

					if (rc) {
						dl_engine_fncs->engine_free(declined_object);
						LOG_ERROR("error, unable to insert declined object in the engine declined object list\n");
					}

				}
				else {
					rc = ENOMEM;
					LOG_ERROR("error, unable to malloc a declined object.\n");
				}
			}
		}
	}


	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Validate the objects in the source and target dlists in the context
 * record.  Remove from the selected objects lists any objects which are no
 * longer acceptable.  Return all acceptable objects in the parameter
 * dlists.  Also, for any object which is removed from the selected dlists,
 * or is otherwise not acceptable, create a declined_handle_t struct with
 * reason why not acceptable and add to the declined_objects dlist.
 */
static int dl_set_shrink_object( task_context_t * context,
                                 dlist_t          declined_objects,
                                 task_effect_t  * effect )
{
	int dlist_rc,rc = EINVAL;
	storage_object_t  *obj=NULL;
	storage_object_t  *child;
	void              *handle;
	boolean            invalidate_objects=FALSE;
	declined_object_t *declined_object=NULL;
	uint               selected_objects_count=0;
	int                search_index=0;
	int                declined_object_count=0;
	int                i;
	drivelink_private_data_t *pdata=NULL;


	LOG_ENTRY();

	REQUIRE(context != NULL);
	REQUIRE(declined_objects != NULL);
	REQUIRE(effect != NULL);
	REQUIRE(context->selected_objects != NULL);
	REQUIRE(context->acceptable_objects != NULL);
	REQUIRE(dl_isa_drivelink(context->object) == TRUE);

 	pdata = (drivelink_private_data_t *) context->object->private_data;

	// reorder the selected objects
	rc = dl_build_reverse_ordered_child_object_list( context->object, context->selected_objects );
	if ( rc ) {
		LOG_ERROR("error, unable to build ordered list of selected objects\n");
		LOG_EXIT_INT(rc);
		return rc;
	}

	GetListSize(context->selected_objects, &selected_objects_count);

	REQUIRE( selected_objects_count > 0 &&
		 selected_objects_count < pdata->drive_link_count );

	declined_object_count = 0;
	search_index          = pdata->drive_link_count - 1;

	DLIST_FOR_EACH(obj, context->selected_objects,dlist_rc ) {

		if (obj != pdata->drive_link[search_index].object) {
			invalidate_objects = TRUE;
		}

		--search_index;

		if ( invalidate_objects == TRUE ) {

			++declined_object_count;

			declined_object = dl_engine_fncs->engine_alloc( sizeof(declined_object_t));

			if (declined_object) {

				declined_object->object = obj;
				declined_object->reason = -1; // EOUTOFORDER;

				// remove object from selectable/acceptable object lists
				DeleteObject( context->selected_objects, obj );
				DeleteObject( context->acceptable_objects, obj );

				// place it on the declined object list
				rc = InsertObject ( declined_objects,
						    declined_object,
						    DECLINED_OBJECT_TAG,
						    NULL,
						    AppendToList,
						    TRUE,
						    &handle );

				if (rc) {
					dl_engine_fncs->engine_free(declined_object);
				}

			}
			else {
				rc = ENOMEM;
				LOG_EXIT_INT(rc);
				return rc;
			}

		}
                
	}
           
	if (declined_object_count == 0) {

		sector_count_t   shrink_sectors=0;

		for( i=pdata->drive_link_count-1; (i>=0)&&(selected_objects_count); i--, selected_objects_count-- ) {

			shrink_sectors += pdata->drive_link[i].sector_count;

			rc = dl_engine_fncs->can_shrink_by( context->object, &shrink_sectors );
			if (rc) {

				shrink_sectors -= pdata->drive_link[i].sector_count;

				++declined_object_count;

				declined_object = dl_engine_fncs->engine_alloc( sizeof(declined_object_t));

				if (declined_object) {

					declined_object->object = pdata->drive_link[i].object;
					declined_object->reason = -1; // ENOCANSHRINK

					// remove object from selectable/acceptable object lists
					DeleteObject( context->selected_objects, obj );
					DeleteObject( context->acceptable_objects, obj );

					// place it on the declined object list
					rc = InsertObject ( declined_objects,
							    declined_object,
							    DECLINED_OBJECT_TAG,
							    NULL,
							    AppendToList,
							    TRUE,
							    &handle );

					if (rc) {
						dl_engine_fncs->engine_free(declined_object);
					}

				}
				else {
					rc = ENOMEM;
					LOG_EXIT_INT(rc);
					return rc;
				}
			}
		}
	}

	// if we can shrink the drivelink Ok then add another link
	// to the acceptable objects list so the user can further
	// shrink the drivelink
	if (declined_object_count == 0) {

		if ( selected_objects_count+1 < pdata->drive_link_count ) {

			child = pdata->drive_link[pdata->drive_link_count-selected_objects_count-1].object;

			rc = ExclusiveInsertObject ( context->acceptable_objects,
						     child,
						     EVMS_OBJECT_TAG,
						     NULL,
						     AppendToList,
						     TRUE,
						     &handle );

			if (rc == 0) {
				GetListSize(context->acceptable_objects, &context->max_selected_objects);
				*effect |= EVMS_Effect_Reload_Objects;
			}
		}
		rc = 0;  // dont stop because we couldn't do the insert here
	}

	// if we declined some objects then we need to fixup the shrink task
	// context so that we only show the good selected objects PLUS the
	// next object that would be Ok to select.

	DeleteAllItems ( context->acceptable_objects, FALSE );
	CopyList(context->acceptable_objects, context->selected_objects, AppendToList );

	GetListSize(context->selected_objects, &context->max_selected_objects);

	if ( context->max_selected_objects+1 < pdata->drive_link_count ) {

		child = pdata->drive_link[pdata->drive_link_count-context->max_selected_objects-1].object;

		rc = ExclusiveInsertObject ( context->acceptable_objects,
					     child,
					     EVMS_OBJECT_TAG,
					     NULL,
					     AppendToList,
					     TRUE,
					     &handle );

		if (rc == 0) {
			++context->max_selected_objects;
		}

	}

	*effect |= EVMS_Effect_Reload_Objects;

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Validate the objects in the selected_objects dlist in the task context.
 * Remove from the selected objects lists any objects which are not
 * acceptable.
 *
 * For unacceptable objects, create a declined_handle_t structure with the
 * reason why it is not acceptable, and add it to the declined_objects dlist.
 * Modify the accepatble_objects dlist in the task context as necessary
 * based on the selected objects and the current settings of the options.
 *
 * Modify any option settings as necessary based on the selected objects.
 * Return the appropriate task_effect_t settings if the object list(s),
 * minimum or maximum objects selected, or option settings have changed.
 */
int dl_set_objects( task_context_t * context,
                   dlist_t          declined_objects,
                   task_effect_t  * effect )
{

	int rc=0;

	LOG_ENTRY();
	
	REQUIRE(context != NULL);
	REQUIRE(declined_objects != NULL);
	REQUIRE(effect != NULL);

	switch (context->action) {

		case EVMS_Task_Create:
			rc = dl_set_create_object( context, declined_objects, effect );
			break;

		case EVMS_Task_Expand:
			rc = dl_set_expand_object( context, declined_objects, effect );
			break;

		case EVMS_Task_Shrink:
			rc = dl_set_shrink_object( context, declined_objects, effect );
			break;
		
		case EVMS_Replace_Missing_DL_Child:
			rc = set_replace_object( context, declined_objects, effect );
			break;

		default:
			rc=EINVAL;
			LOG_ERROR("error, context->action is unknown or unsupported\n");
			break;
        }

	LOG_EXIT_INT(rc);
	return rc;
}


static int dl_set_create_option( task_context_t * context,
                                 u_int32_t        index,
                                 value_t        * value,
                                 u_int32_t      * info )
{
	int rc = EINVAL;

	LOG_ENTRY();

	REQUIRE(context != NULL);
	REQUIRE(index == DL_CREATE_LINKNAME_INDEX);
	REQUIRE(value != NULL);
	REQUIRE(value->s != NULL);
        REQUIRE(strlen(value->s) > 0 &&
                strlen(value->s) <= EVMS_VOLUME_NAME_SIZE );

	rc = dl_engine_fncs->validate_name( value->s );
	
	if (rc == 0) {
		strcpy( context->option_descriptors->option[index].value.s, value->s);
        }

	LOG_EXIT_INT(rc);
	return rc;
}


int dl_set_option( task_context_t * context,
                   u_int32_t        index,
                   value_t        * value,
                   u_int32_t      * info )
{
	int rc=EINVAL;

	LOG_ENTRY();
	
	REQUIRE(context != NULL);
	REQUIRE(index == DL_CREATE_LINKNAME_INDEX);
	REQUIRE(value != NULL);
	REQUIRE(value->s != NULL);

        switch (context->action) {

		case EVMS_Task_Create:
		case EVMS_Task_Add_Feature:
			rc = dl_set_create_option( context, index, value, info );
			break;

 		default:
			rc = ENOSYS;
			break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Validate the objects in the source and target dlists in the context
 * record.  Remove from the selected objects lists any objects which are no
 * longer acceptable.  Return all acceptable objects in the parameter
 * dlists.  Also, for any object which is removed from the selected dlists,
 * or is otherwise not acceptable, create a declined_handle_t struct with
 * reason why not acceptable and add to the declined_objects dlist.
 */
static int dl_get_acceptable_create_objects( task_context_t * context )
{
	int   rc = EINVAL;
	uint  count;

	LOG_ENTRY();

	REQUIRE(context != NULL);

	GetListSize(context->acceptable_objects, &count);

	REQUIRE(count == 0); // should not have any acceptable objects yet

        rc = dl_engine_fncs->get_object_list( 0,
					      DATA_TYPE,
					      NULL,
					      NULL,
					      VALID_INPUT_OBJECT,
					      &context->acceptable_objects );

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 *  Called from get_acceptable_expand_objects
 */
static boolean dl_prune_expand_object_list( ADDRESS   Object,
					    TAG       ObjectTag,
					    ADDRESS   ObjectHandle,
					    ADDRESS   Parameters,
					    boolean  *FreeMemory,
					    uint     *Error )
{
	storage_object_t * obj = (storage_object_t *) Object;
	storage_object_t * expand_object = (storage_object_t *)Parameters;

	*FreeMemory = FALSE;           // tells dlist not to free any memory
	*Error      = DLIST_SUCCESS;   // tells dlist we were successful

	if ( obj != expand_object &&
	     obj->disk_group == expand_object->disk_group &&
	     dl_isa_parent(obj, expand_object) == FALSE ) {
		return FALSE;
	}

	return TRUE;           // tells dlist to prune the object from the list
}


/*
 * Validate the objects in the source and target dlists in the context
 * record.  Remove from the selected objects lists any objects which are no
 * longer acceptable.  Return all acceptable objects in the parameter
 * dlists.  Also, for any object which is removed from the selected dlists,
 * or is otherwise not acceptable, create a declined_handle_t struct with
 * reason why not acceptable and add to the declined_objects dlist.
 */
static int dl_get_acceptable_expand_objects( task_context_t * context )
{
	int   rc = EINVAL;

	LOG_ENTRY();

	REQUIRE(context != NULL);

	LOG_DEBUG("getting expand objects for drivelink %s\n", context->object->name);

        rc = dl_engine_fncs->get_object_list( 0,
					      DATA_TYPE,
					      NULL,
					      NULL,
					      VALID_INPUT_OBJECT,
					      &context->acceptable_objects );

        if (rc == 0) {
		PruneList(context->acceptable_objects, dl_prune_expand_object_list, context->object);
        }

	LOG_EXIT_INT(rc);
	return rc;
}



/*
 * Validate the objects in the source and target dlists in the context
 * record.  Remove from the selected objects lists any objects which are no
 * longer acceptable.  Return all acceptable objects in the parameter
 * dlists.  Also, for any object which is removed from the selected dlists,
 * or is otherwise not acceptable, create a declined_handle_t struct with
 * reason why not acceptable and add to the declined_objects dlist.
 */
static int dl_get_acceptable_shrink_objects( task_context_t * context  )
{
	int                       rc=0;
	void                     *handle;
	drivelink_private_data_t *pdata;
	storage_object_t         *child;


	LOG_ENTRY();

	REQUIRE(context != NULL);
	REQUIRE(dl_isa_drivelink(context->object) == TRUE);
	REQUIRE(context->acceptable_objects != NULL);

	LOG_DEBUG("getting shrink object for drivelink %s\n", context->object->name);

        pdata = (drivelink_private_data_t *) context->object->private_data;

        if (pdata->drive_link_count > 1) {

		child = pdata->drive_link[pdata->drive_link_count-1].object;

                rc = InsertObject ( context->acceptable_objects,
                                    child,
                                    EVMS_OBJECT_TAG,
                                    NULL,
                                    AppendToList,
                                    TRUE,
                                    &handle );
        }

	LOG_EXIT_INT(rc);
	return rc;
}



/*
 * Validate the objects in the source and target dlists in the context
 * record.  Remove from the selected objects lists any objects which are no
 * longer acceptable.  Return all acceptable objects in the parameter
 * dlists.  Also, for any object which is removed from the selected dlists,
 * or is otherwise not acceptable, create a declined_handle_t struct with
 * reason why not acceptable and add to the declined_objects dlist.
 */
int dl_init_task( task_context_t * context )
{
	int rc = EINVAL;
	drivelink_private_data_t *pdata=NULL;

	LOG_ENTRY();

	REQUIRE(context != NULL);
       
        switch (context->action) {

		case EVMS_Task_Create:

			context->min_selected_objects = 1;
			context->max_selected_objects = EVMS_DRIVELINK_MAX_ENTRIES;

			rc = dl_allocate_create_option_descriptors( context );
			if (rc == 0) {
				rc = dl_get_acceptable_create_objects( context );
			}
			break;

		case EVMS_Task_Expand:
			
			REQUIRE(dl_isa_drivelink(context->object) == TRUE);
			
			pdata = (drivelink_private_data_t *) context->object->private_data;

			context->min_selected_objects = 1;
			context->max_selected_objects = EVMS_DRIVELINK_MAX_ENTRIES - pdata->drive_link_count;

			rc = dl_allocate_expand_option_descriptors( context );
			if (rc == 0) {
				rc = dl_get_acceptable_expand_objects( context );
			}
			break;

		case EVMS_Task_Shrink:

			REQUIRE(dl_isa_drivelink(context->object) == TRUE);

			context->min_selected_objects = 1;

			rc = dl_allocate_shrink_option_descriptors( context );
			if (rc == 0) {
				rc = dl_get_acceptable_shrink_objects( context );
				if (rc ==0) {
					GetListSize(context->acceptable_objects, &context->max_selected_objects);
				}
			}
			break;

		case EVMS_Task_Add_Feature:
			rc = dl_allocate_create_option_descriptors( context );
			if (rc == 0) {
				rc = dl_initialize_create_option_descriptors( context );
			}
			break;

		case EVMS_Replace_Missing_DL_Child:
			context->min_selected_objects = 1;
			context->max_selected_objects = 1;
			rc = allocate_replace_option_descriptors( context );
			if (rc == 0) {
				rc = get_acceptable_replace_objects( context );
			}
			break;

		default:
			LOG_ERROR("context->action is unknown or unsupported\n");
			break;
	}


	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Returns specific information
 */
int dl_get_info( storage_object_t  * object, char *name,  extended_info_array_t * * info)
{
	int rc = EINVAL;
	extended_info_array_t    *Info;
	drivelink_private_data_t  *pdata;

	LOG_ENTRY();

	REQUIRE(dl_isa_drivelink(object) == TRUE);
	REQUIRE(info != NULL);
	
	pdata = (drivelink_private_data_t *) object->private_data;

        Info = dl_engine_fncs->engine_alloc( sizeof(extended_info_array_t) + ( DL_INFO_COUNT * sizeof(extended_info_t) ) );
        if (Info) {

		// the following info isnt available for missing child objects
		if (pdata->signature & MISSING_CHILD_SIGNATURE) {

			SET_STRING( Info->info[DL_INFO_NAME_INDEX].name, "Name" );
			SET_STRING( Info->info[DL_INFO_NAME_INDEX].title, "Name" );
			SET_STRING( Info->info[DL_INFO_NAME_INDEX].desc, "A temporary name given to the missing drivelink object that is used to assist with recovery.");
			Info->info[DL_INFO_NAME_INDEX].type               = EVMS_Type_String;
			Info->info[DL_INFO_NAME_INDEX].unit               = EVMS_Unit_None;
			SET_STRING( Info->info[DL_INFO_NAME_INDEX].value.s, object->name );
			Info->info[DL_INFO_NAME_INDEX].collection_type    = EVMS_Collection_None;
			memset( &Info->info[DL_INFO_NAME_INDEX].group, 0, sizeof(group_info_t));

			Info->count += 1;

			SET_STRING( Info->info[DL_INFO_SIZE_INDEX].name, "Size" );
			SET_STRING( Info->info[DL_INFO_SIZE_INDEX].title, "Size" );
			SET_STRING( Info->info[DL_INFO_SIZE_INDEX].desc, "The size of the object that is missing, i.e., the useable size after subtracting metadata.");
			Info->info[DL_INFO_SIZE_INDEX].type               = EVMS_Type_Unsigned_Int64;
			Info->info[DL_INFO_SIZE_INDEX].unit               = EVMS_Unit_Sectors;
			Info->info[DL_INFO_SIZE_INDEX].value.ui64         = object->size;
			Info->info[DL_INFO_SIZE_INDEX].collection_type    = EVMS_Collection_None;
			memset( &Info->info[DL_INFO_SIZE_INDEX].group, 0, sizeof(group_info_t));
			Info->info[DL_INFO_SIZE_INDEX].flags |= EVMS_EINFO_FLAGS_NO_UNIT_CONVERSION;

			Info->count += 1;

		}
		else {

			SET_STRING( Info->info[DL_INFO_NAME_INDEX].name, "Name" );
			SET_STRING( Info->info[DL_INFO_NAME_INDEX].title, "Name" );
			SET_STRING( Info->info[DL_INFO_NAME_INDEX].desc, "This is the name given to the storage object. It must be unique on the system.");
			Info->info[DL_INFO_NAME_INDEX].type               = EVMS_Type_String;
			Info->info[DL_INFO_NAME_INDEX].unit               = EVMS_Unit_None;
			SET_STRING( Info->info[DL_INFO_NAME_INDEX].value.s, object->name );
			Info->info[DL_INFO_NAME_INDEX].collection_type    = EVMS_Collection_None;
			memset( &Info->info[DL_INFO_NAME_INDEX].group, 0, sizeof(group_info_t));

			Info->count += 1;

			SET_STRING( Info->info[DL_INFO_SIZE_INDEX].name, "Size" );
			SET_STRING( Info->info[DL_INFO_SIZE_INDEX].title, "Size" );
			SET_STRING( Info->info[DL_INFO_SIZE_INDEX].desc, "This is the size of the entire drivelink, i.e., the sum of all the drivelink child storage objects.");
			Info->info[DL_INFO_SIZE_INDEX].type               = EVMS_Type_Unsigned_Int64;
			Info->info[DL_INFO_SIZE_INDEX].unit               = EVMS_Unit_Sectors;
			Info->info[DL_INFO_SIZE_INDEX].value.ui64         = object->size;
			Info->info[DL_INFO_SIZE_INDEX].collection_type    = EVMS_Collection_None;
			memset( &Info->info[DL_INFO_SIZE_INDEX].group, 0, sizeof(group_info_t));
			Info->info[DL_INFO_SIZE_INDEX].flags |= EVMS_EINFO_FLAGS_NO_UNIT_CONVERSION;

			Info->count += 1;

			SET_STRING( Info->info[DL_INFO_LINKS_INDEX].name, "Links" );
			SET_STRING( Info->info[DL_INFO_LINKS_INDEX].title, "Children" );
			SET_STRING( Info->info[DL_INFO_LINKS_INDEX].desc, "This is the number of child storage objects being linked by the drivelink feature.");
			Info->info[DL_INFO_LINKS_INDEX].type               = EVMS_Type_Unsigned_Int64;
			Info->info[DL_INFO_LINKS_INDEX].unit               = EVMS_Unit_None;
			Info->info[DL_INFO_LINKS_INDEX].value.ui64         = pdata->drive_link_count;
			Info->info[DL_INFO_LINKS_INDEX].collection_type    = EVMS_Collection_None;
			memset( &Info->info[DL_INFO_LINKS_INDEX].group, 0, sizeof(group_info_t));

			Info->count += 1;

			SET_STRING( Info->info[DL_INFO_PSN_INDEX].name, "PSN" );
			SET_STRING( Info->info[DL_INFO_PSN_INDEX].title, "Parent SN" );
			SET_STRING( Info->info[DL_INFO_PSN_INDEX].desc, "The serial number of the parent drivelink object.");
			Info->info[DL_INFO_PSN_INDEX].type               = EVMS_Type_Unsigned_Int64;
			Info->info[DL_INFO_PSN_INDEX].unit               = EVMS_Unit_None;
			Info->info[DL_INFO_PSN_INDEX].format             = EVMS_Format_Hex;
			Info->info[DL_INFO_PSN_INDEX].flags              = EVMS_OPTION_FLAGS_NO_UNIT_CONVERSION;
			Info->info[DL_INFO_PSN_INDEX].value.ui64         = pdata->parent_serial_number;
			Info->info[DL_INFO_PSN_INDEX].collection_type    = EVMS_Collection_None;
			memset( &Info->info[DL_INFO_PSN_INDEX].group, 0, sizeof(group_info_t));

			Info->count += 1;

		}


		*info = Info;
		rc = 0;
	}
        else {
		rc = ENOMEM;
		LOG_ERROR("unable to malloc memory for extended info array\n");
        }
    

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 *
 */
int dl_get_plugin_info( char * descriptor_name, extended_info_array_t * * info )
{
	int rc = EINVAL;
	extended_info_array_t   *Info;
	char                     version_string[64];
	char                     required_engine_api_version_string[64];
	char                     required_plugin_api_version_string[64];


	LOG_ENTRY();

	REQUIRE(info != NULL);

	*info = NULL;     // init to no info returned

	Info = dl_engine_fncs->engine_alloc( sizeof(extended_info_array_t) + (DL_PLUGIN_INFO_COUNT*sizeof(extended_info_t))  );
	if (Info) {

		Info->count = DL_PLUGIN_INFO_COUNT;

		sprintf(version_string, "%d.%d.%d",
			MAJOR_VERSION,
			MINOR_VERSION,
			PATCH_LEVEL );

		sprintf(required_engine_api_version_string, "%d.%d.%d",
			dl_plugin_record->required_engine_api_version.major,
			dl_plugin_record->required_engine_api_version.minor,
			dl_plugin_record->required_engine_api_version.patchlevel );

		sprintf(required_plugin_api_version_string, "%d.%d.%d",
			dl_plugin_record->required_plugin_api_version.plugin.major,
			dl_plugin_record->required_plugin_api_version.plugin.minor,
			dl_plugin_record->required_plugin_api_version.plugin.patchlevel );

		SET_STRING( Info->info[DL_PLUGIN_INFO_SNAME_INDEX].name, "Short Name" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_SNAME_INDEX].title, "Short Name" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_SNAME_INDEX].desc, "A short name given to this plugin.");
		Info->info[DL_PLUGIN_INFO_SNAME_INDEX].type               = EVMS_Type_String;
		Info->info[DL_PLUGIN_INFO_SNAME_INDEX].unit               = EVMS_Unit_None;
		SET_STRING( Info->info[DL_PLUGIN_INFO_SNAME_INDEX].value.s, dl_plugin_record->short_name );
		Info->info[DL_PLUGIN_INFO_SNAME_INDEX].collection_type    = EVMS_Collection_None;
		memset( &Info->info[DL_PLUGIN_INFO_SNAME_INDEX].group, 0, sizeof(group_info_t));

		SET_STRING( Info->info[DL_PLUGIN_INFO_LNAME_INDEX].name, "Long Name" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_LNAME_INDEX].title, "Long Name" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_LNAME_INDEX].desc, "A long name given to this plugin.");
		Info->info[DL_PLUGIN_INFO_LNAME_INDEX].type               = EVMS_Type_String;
		Info->info[DL_PLUGIN_INFO_LNAME_INDEX].unit               = EVMS_Unit_None;
		SET_STRING( Info->info[DL_PLUGIN_INFO_LNAME_INDEX].value.s, dl_plugin_record->long_name );
		Info->info[DL_PLUGIN_INFO_LNAME_INDEX].collection_type    = EVMS_Collection_None;
		memset( &Info->info[DL_PLUGIN_INFO_LNAME_INDEX].group, 0, sizeof(group_info_t));

		SET_STRING( Info->info[DL_PLUGIN_INFO_TYPE_INDEX].name, "Type" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_TYPE_INDEX].title, "Plugin Type" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_TYPE_INDEX].desc, "There are various types of plugins; each responsible for some kind of storage object.");
		Info->info[DL_PLUGIN_INFO_TYPE_INDEX].type               = EVMS_Type_String;
		Info->info[DL_PLUGIN_INFO_TYPE_INDEX].unit               = EVMS_Unit_None;
		SET_STRING( Info->info[DL_PLUGIN_INFO_TYPE_INDEX].value.s, "Aggregate Class Feature" );
		Info->info[DL_PLUGIN_INFO_TYPE_INDEX].collection_type    = EVMS_Collection_None;
		memset( &Info->info[DL_PLUGIN_INFO_TYPE_INDEX].group, 0, sizeof(group_info_t));

		SET_STRING( Info->info[DL_PLUGIN_INFO_VERSION_INDEX].name, "Version" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_VERSION_INDEX].title, "Plugin Version" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_VERSION_INDEX].desc, "This is the version number of the plugin.");
		Info->info[DL_PLUGIN_INFO_VERSION_INDEX].type               = EVMS_Type_String;
		Info->info[DL_PLUGIN_INFO_VERSION_INDEX].unit               = EVMS_Unit_None;
		SET_STRING( Info->info[DL_PLUGIN_INFO_VERSION_INDEX].value.s, version_string );
		Info->info[DL_PLUGIN_INFO_VERSION_INDEX].collection_type    = EVMS_Collection_None;
		memset( &Info->info[DL_PLUGIN_INFO_VERSION_INDEX].group, 0, sizeof(group_info_t));

		SET_STRING( Info->info[DL_PLUGIN_INFO_REQENGSVCVERSION_INDEX].name, "Required Engine Services Version" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_REQENGSVCVERSION_INDEX].title, "Required Engine Services Version" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_REQENGSVCVERSION_INDEX].desc, "This is the version of the Engine services that this plug-in requires. It will not run on older versions of the Engine services.");
		Info->info[DL_PLUGIN_INFO_REQENGSVCVERSION_INDEX].type               = EVMS_Type_String;
		Info->info[DL_PLUGIN_INFO_REQENGSVCVERSION_INDEX].unit               = EVMS_Unit_None;
		SET_STRING( Info->info[DL_PLUGIN_INFO_REQENGSVCVERSION_INDEX].value.s, required_engine_api_version_string );
		Info->info[DL_PLUGIN_INFO_REQENGSVCVERSION_INDEX].collection_type    = EVMS_Collection_None;
		memset( &Info->info[DL_PLUGIN_INFO_REQENGSVCVERSION_INDEX].group, 0, sizeof(group_info_t));

		SET_STRING( Info->info[DL_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].name, "Required Engine Plug-in API Version" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].title, "Required Engine Plug-in API Version" );
		SET_STRING( Info->info[DL_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].desc, "This is the version of the Engine plug-in API that this plug-in requires. It will not run on older versions of the Engine plug-in API.");
		Info->info[DL_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].type               = EVMS_Type_String;
		Info->info[DL_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].unit               = EVMS_Unit_None;
		SET_STRING( Info->info[DL_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].value.s, required_plugin_api_version_string );
		Info->info[DL_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].collection_type    = EVMS_Collection_None;
		memset( &Info->info[DL_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].group, 0, sizeof(group_info_t));

		*info = Info;
		rc=0;
	}


	LOG_EXIT_INT(rc);
	return rc;
}

