/*
 *   (C) Copyright IBM Corp. 2001, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: LVM Region Manager
 * File: evms2/engine/plugins/lvm/lvm_dlist.c
 *
 * Description: This file contains all functions related to manipulating the
 *              dlists that are used by the LVM region manager.
 */ 

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "lvmregmgr.h"

/**
 * lvm_append_segment_to_container
 *
 * Append the specified new_segment to the end of the "consumed" dlist in
 * the specified container.
 **/
int lvm_append_segment_to_container(storage_object_t * new_segment,
				    storage_container_t * container)
{
	int rc;
	void * handle;

	LOG_ENTRY();

	rc = InsertObject(container->objects_consumed, new_segment,
			  get_tag_for_object(new_segment), NULL, AppendToList,
			  TRUE, &handle);
	if (!rc) {
		/* Mark the segment as belonging to this container. */
		new_segment->consuming_container = container;
	} else {
		LOG_SERIOUS("Error adding object %s to container %s\n",
			    new_segment->name, container->name );
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 * lvm_remove_segment_from_container
 *
 * Remove the specified segment from its "consuming" container.
 **/
int lvm_remove_segment_from_container(storage_object_t * segment)
{
	int rc = EINVAL;

	LOG_ENTRY();

	if (!segment->consuming_container) {
		LOG_WARNING("Object %s is not in a container\n", segment->name);
		goto out;
	}

	rc = DeleteObject(segment->consuming_container->objects_consumed,
			  segment);
	if (!rc) {
		/* Mark the segment as "unconsumed". */
		segment->consuming_container = NULL;
	} else {
		LOG_SERIOUS("Error removing object %s from container %s\n",
			    segment->name, segment->consuming_container->name);
	}

out:
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 * lvm_append_region_to_container
 *
 * Append the specified new_region at the end of the "producing" dlist in
 * the specified container.
 **/
int lvm_append_region_to_container(storage_object_t * new_region,
				   storage_container_t * container)
{
	int rc;
	void * handle;

	LOG_ENTRY();

	rc = InsertObject(container->objects_produced, new_region,
			  get_tag_for_object(new_region), NULL,
			  AppendToList, TRUE, &handle);
	if (!rc) {
		/* Mark the region as belonging to this container. */
		new_region->producing_container = container;
	} else {
		LOG_SERIOUS("Error adding region %s to container %s\n",
			    new_region->name, container->name);
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 * lvm_remove_region_from_container
 *
 * Remove the specified region from its "producing" container.
 **/
int lvm_remove_region_from_container(storage_object_t * region)
{
	int rc = EINVAL;

	LOG_ENTRY();

	if (!region->producing_container) {
		LOG_WARNING("Region %s not in a container\n", region->name);
		goto out;
	}

	rc = DeleteObject(region->producing_container->objects_produced, region);
	if (!rc) {
		/* Mark the region as "unproduced". */
		region->producing_container = NULL;
	} else {
		LOG_SERIOUS("Error removing region %s from container %s\n",
			    region->name, region->producing_container->name);
	}

out:
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 * lvm_append_region_to_segment
 *
 * Associate the specified region and segment as parent/child. Add the
 * region to the segment's "parent" dlist, and add the segment to the
 * region's "child" dlist. Also need to check for duplicates, so the
 * same region/segment don't get associated more than once.
 **/
int lvm_append_region_to_segment(storage_object_t * region,
				 storage_object_t * segment)
{
	int rc;
	void * handle;

	/* No LOG_ENTRY or LOG_EXIT calls. Writes too many messages to the log. */

	rc = ExclusiveInsertObject(segment->parent_objects, region,
				   get_tag_for_object(region), NULL,
				   AppendToList, TRUE, &handle);
	if (rc) {
		LOG_SERIOUS("Error adding region %s as a parent to object %s\n",
			    region->name, segment->name);
		return rc;
	}

	rc = ExclusiveInsertObject(region->child_objects, segment,
				   get_tag_for_object(segment), NULL,
				   AppendToList, TRUE, &handle);
	if (rc) {
		LOG_SERIOUS("Error adding object %s as a child to region %s\n",
			    segment->name, region->name);
		DeleteObject(segment->parent_objects, region);
	}

	return rc;
}

/**
 * lvm_remove_region_from_segment
 *
 * Remove the parent/child association between the specified region
 * and segment. Remove the region from the segment's "parent" list,
 * and remove the segment from the region's "child" list.
 **/
int lvm_remove_region_from_segment(storage_object_t * region,
				   storage_object_t * segment)
{
	int rc;

	LOG_ENTRY();

	rc = DeleteObject(segment->parent_objects, region);
	if (rc) {
		LOG_SERIOUS("Error removing region %s from object %s\n",
			    region->name, segment->name);
	}

	rc = DeleteObject(region->child_objects, segment);
	if (rc) {
		LOG_SERIOUS("Error removing object %s from region %s\n",
			    segment->name, region->name);
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 * lvm_clear_child_list
 *
 * Remove all segments from this region's child list.
 **/
int lvm_clear_child_list(storage_object_t * region)
{
	storage_object_t * segment = NULL;
	TAG tag;

	LOG_ENTRY();

	GoToStartOfList(region->child_objects);

	while (! BlindGetObject(region->child_objects, &tag,
				NULL, FALSE, (void**)&segment)) {
		if (segment) {
			if (lvm_remove_region_from_segment(region, segment)) {
				LOG_SERIOUS("Could not clear all objects from child list in region %s\n",
					    region->name);
			}
		}
	}

	LOG_EXIT_INT(0);
	return 0;
}

/**
 * lvm_add_object_to_list
 *
 * Add the specified object to the specified list.
 **/
int lvm_add_object_to_list(storage_object_t * object,
			   dlist_t list)
{
	int rc;
	void * handle;

	LOG_ENTRY();

	rc = InsertObject(list, object, get_tag_for_object(object),
			  NULL, AppendToList, TRUE, &handle);
	if (rc) {
		LOG_SERIOUS("Error adding object %s to list\n", object->name);
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 * lvm_add_group_to_list
 *
 * Add the specified group to the global list.
 **/
int lvm_add_group_to_list(lvm_volume_group_t * group)
{
	void * handle;
	int rc;

	LOG_ENTRY();

	rc = InsertObject(lvm_group_list, group, 0, NULL,
			  AppendToList, TRUE, &handle);
	if (rc) {
		LOG_SERIOUS("Error adding container %s to global list\n",
			    group->container->name);
	}

	LOG_EXIT_INT(rc);
	return rc;
}

